"""
<name>GO Browser</name>
<description>Enrichment analysis for Gene Ontology terms.</description>
<contact>Ales Erjavec</contact>
<icon>icons/GOTermFinder.png</icon>
<priority>2020</priority>
"""

from __future__ import absolute_import, with_statement

from collections import defaultdict
from functools import partial
import gc
import sys, os, tarfile, math
from os.path import join as p_join

from Orange.orng import orngServerFiles
from Orange.orng.orngDataCaching import data_hints
from Orange.OrangeWidgets import OWGUI
from Orange.OrangeWidgets.OWWidget import *

from .. import obiGene, obiGO, obiProb, obiTaxonomy

dataDir = orngServerFiles.localpath("GO")

def listAvailable():
    files = orngServerFiles.listfiles("GO")
    ret = {}
    for file in files:
        tags = orngServerFiles.info("GO", file)["tags"]
        td = dict([tuple(tag.split(":")) for tag in tags if tag.startswith("#") and ":" in tag])
        if "association" in file.lower():
            ret[td.get("#organism", file)] = file
    orgMap = {"352472":"44689"}
    essential = ["gene_association.%s.tar.gz" % obiGO.from_taxid(id) for id in obiTaxonomy.essential_taxids() if obiGO.from_taxid(id)]
    essentialNames = [obiTaxonomy.name(id) for id in obiTaxonomy.essential_taxids() if obiGO.from_taxid(id)]
    ret.update(zip(essentialNames, essential))
    return ret

class _disablegc(object):
    def __enter__(self):
        gc.disable()
    def __exit__(self, *args):
        gc.enable()

def getOrgFileName(org):
    from Orange.orng import orngServerFiles
    files = orngServerFiles.listfiles("go")
    return [f for f in files if org in f].pop()

class TreeNode(object):
    def __init__(self, tuple, children):
        self.tuple = tuple
        self.children = children

class GOTreeWidget(QTreeWidget):
    def contextMenuEvent(self, event):
        QTreeWidget.contextMenuEvent(self, event)
##        print event.x(), event.y()
        term = self.itemAt(event.pos()).term
        self._currMenu = QMenu()
        self._currAction = self._currMenu.addAction("View term on AmiGO website")
##        self.connect(self, SIGNAL("triggered(QAction*)"), partial(self.BrowserAction, term))
        self.connect(self._currAction, SIGNAL("triggered()"), lambda :self.BrowserAction(term))
        self._currMenu.popup(event.globalPos())

    def BrowserAction(self, term):
        import webbrowser
        if isinstance(term, obiGO.Term):
            term = term.id
        webbrowser.open("http://amigo.geneontology.org/cgi-bin/amigo/term-details.cgi?term="+term)
        
    def paintEvent(self, event):
        QTreeWidget.paintEvent(self, event)
        if getattr(self, "_userMessage", None):
            painter = QPainter(self.viewport())
            font = QFont(self.font())
            font.setPointSize(15)
            painter.setFont(font)
            painter.drawText(self.viewport().geometry(), Qt.AlignCenter, self._userMessage)
            painter.end()

class OWGOEnrichmentAnalysis(OWWidget):
    settingsList=["annotationIndex", "useReferenceDataset", "aspectIndex", "geneAttrIndex", "geneMatcherSettings",
                    "filterByNumOfInstances", "minNumOfInstances", "filterByPValue", "maxPValue", "selectionDirectAnnotation", "selectionDisjoint", "selectionType",
                    "selectionAddTermAsClass", "useAttrNames", "probFunc", "useFDR"]
    contextHandlers = {"": DomainContextHandler("", ["geneAttrIndex", "useAttrNames", "annotationIndex", "geneMatcherSettings"], matchValues=1)}
    def __init__(self, parent=None, signalManager=None, name="GO Browser"):
        OWWidget.__init__(self, parent, signalManager, name)
        self.inputs = [("Cluster Examples", ExampleTable, self.SetClusterDataset, Default), ("Reference Examples", ExampleTable, self.SetReferenceDataset, Single + NonDefault)] #, ("Structured Data", DataFiles, self.chipdata, Single + NonDefault)]
        self.outputs = [("Selected Examples", ExampleTable, Default), ("Unselected Examples", ExampleTable, Default), ("Example With Unknown Genes", ExampleTable, Default)] #, ("Selected Structured Data", DataFiles, Single + NonDefault)]

        self.annotationIndex = 0
        self.autoFindBestOrg = False
        self.useReferenceDataset  = 0
        self.aspectIndex = 0
        self.geneAttrIndex = 0
        self.useAttrNames = False
        self.geneMatcherSettings = [True, False, False, False]
        self.filterByNumOfInstances = False
        self.minNumOfInstances = 1
        self.filterByPValue = True
        self.maxPValue = 0.1
        self.probFunc = 0
        self.useFDR = True
        self.selectionDirectAnnotation = 0
        self.selectionDisjoint = 0
        self.selectionAddTermAsClass = 0
        self.selectionChanging = 0
        
        # check usage of all evidences
        for etype in obiGO.evidenceTypesOrdered:
            varName = "useEvidence" + etype
            if varName not in self.settingsList: 
                self.settingsList.append(varName)
            code = compile("self.%s = True" % (varName), ".", "single")
            exec(code)
        self.annotationCodes = []
        
        self.loadSettings()
        
        #############
        ##GUI
        #############
        self.tabs = OWGUI.tabWidget(self.controlArea)
        ##Input tab
        self.inputTab = OWGUI.createTabPage(self.tabs, "Input")
        box = OWGUI.widgetBox(self.inputTab, "Info")
        self.infoLabel = OWGUI.widgetLabel(box, "No data on input\n")
        
        OWGUI.button(box, self, "Ontology/Annotation Info", callback=self.ShowInfo,
                     tooltip="Show information on loaded ontology and annotations",
                     debuggingEnabled=0)
        box = OWGUI.widgetBox(self.inputTab, "Organism", addSpace=True)
        self.annotationComboBox = OWGUI.comboBox(box, self, "annotationIndex",
                            items = self.annotationCodes, callback=self.Update,
                            tooltip="Select organism", debuggingEnabled=0)
        
        # freeze until annotation combo box is updateded with available annotations.
        self.signalManager.freeze(self).push()
        QTimer.singleShot(0, self.UpdateOrganismComboBox)
        
        self.geneAttrIndexCombo = OWGUI.comboBox(self.inputTab, self, "geneAttrIndex",
                            box="Gene names", callback=self.Update,
                            tooltip="Use this attribute to extract gene names from input data")
        OWGUI.checkBox(self.geneAttrIndexCombo.box, self, "useAttrNames", "Use data attributes names",
                       disables=[(-1, self.geneAttrIndexCombo)], callback=self.Update, 
                       tooltip="Use attribute names for gene names")
        OWGUI.button(self.geneAttrIndexCombo.box, self, "Gene matcher settings", 
                     callback=self.UpdateGeneMatcher, 
                     tooltip="Open gene matching settings dialog", 
                     debuggingEnabled=0)
        
        self.referenceRadioBox = OWGUI.radioButtonsInBox(self.inputTab, self, "useReferenceDataset", 
                                                         ["Entire genome", "Reference set (input)"],
                                                         tooltips=["Use entire genome for reference",
                                                                   "Use genes from Referece Examples input signal as reference"],
                                                         box="Reference", callback=self.Update)
        self.referenceRadioBox.buttons[1].setDisabled(True)
        OWGUI.radioButtonsInBox(self.inputTab, self, "aspectIndex", ["Biological process",
                                                                     "Cellular component",
                                                                     "Molecular function"], 
                                box="Aspect", callback=self.Update)
        
        self.geneAttrIndexCombo.setDisabled(bool(self.useAttrNames))
        
        ##Filter tab
        self.filterTab = OWGUI.createTabPage(self.tabs, "Filter")
        box = OWGUI.widgetBox(self.filterTab, "Filter GO Term Nodes", addSpace=True)
        OWGUI.checkBox(box, self, "filterByNumOfInstances", "Genes",
                       callback=self.FilterAndDisplayGraph, 
                       tooltip="Filter by number of input genes mapped to a term")
        OWGUI.spin(OWGUI.indentedBox(box), self, 'minNumOfInstances', 1, 100, 
                   step=1, label='#:', labelWidth=15, 
                   callback=self.FilterAndDisplayGraph, 
                   callbackOnReturn=True, 
                   tooltip="Min. number of input genes mapped to a term")
        
        OWGUI.checkBox(box, self, "filterByPValue", "Significance",
                       callback=self.FilterAndDisplayGraph, 
                       tooltip="Filter by term p-value")
        OWGUI.doubleSpin(OWGUI.indentedBox(box), self, 'maxPValue', 1e-8, 1, 
                         step=1e-8,  label='p:', labelWidth=15, 
                         callback=self.FilterAndDisplayGraph, 
                         callbackOnReturn=True, 
                         tooltip="Max term p-value")
        box = OWGUI.widgetBox(box, "Significance test")
        OWGUI.radioButtonsInBox(box, self, "probFunc", ["Binomial", "Hypergeometric"], 
                                tooltips=["Use binomial distribution test", 
                                          "Use hypergeometric distribution test"], 
                                callback=self.Update)
        OWGUI.checkBox(box, self, "useFDR", "Use FDR (False Discovery Rate)", 
                       callback=self.Update, 
                       tooltip="Use False Discovery Rate correction")
        box = OWGUI.widgetBox(self.filterTab, "Evidence codes in annotation", 
                              addSpace=True)
        self.evidenceCheckBoxDict = {}
        for etype in obiGO.evidenceTypesOrdered:
            self.evidenceCheckBoxDict[etype] = OWGUI.checkBox(box, self, "useEvidence"+etype, etype,
                                            callback=self.Update, tooltip=obiGO.evidenceTypes[etype])
        
        ##Select tab
        self.selectTab = OWGUI.createTabPage(self.tabs, "Select")
        box = OWGUI.radioButtonsInBox(self.selectTab, self, "selectionDirectAnnotation", 
                                      ["Directly or Indirectly", "Directly"], 
                                      box="Annotated genes", 
                                      callback=self.ExampleSelection)
        
        box = OWGUI.widgetBox(self.selectTab, "Output", addSpace=True)
        OWGUI.radioButtonsInBox(box, self, "selectionDisjoint", 
                                btnLabels=["All selected genes", 
                                           "Term-specific genes", 
                                           "Common term genes"], 
                                tooltips=["Outputs genes annotated to all selected GO terms", 
                                          "Outputs genes that appear in only one of selected GO terms", 
                                          "Outputs genes common to all selected GO terms"], 
                                callback=[self.ExampleSelection,
                                          self.UpdateAddClassButton])
        self.addClassCB = OWGUI.checkBox(box, self, "selectionAddTermAsClass",
                                         "Add GO Term as class", 
                                         callback=self.ExampleSelection)

        # ListView for DAG, and table for significant GOIDs
        self.DAGcolumns = ['GO term', 'Cluster', 'Reference', 'p value', 'Genes', 'Enrichment']
        
        self.splitter = QSplitter(Qt.Vertical, self.mainArea)
        self.mainArea.layout().addWidget(self.splitter)

        # list view
        self.listView = GOTreeWidget(self.splitter)
        self.listView.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.listView.setAllColumnsShowFocus(1)
        self.listView.setColumnCount(len(self.DAGcolumns))
        self.listView.setHeaderLabels(self.DAGcolumns)
        
        self.listView.header().setClickable(True)
        self.listView.header().setSortIndicatorShown(True)
        self.listView.setSortingEnabled(True)
        self.listView.setItemDelegateForColumn(5, EnrichmentColumnItemDelegate(self))
        self.listView.setRootIsDecorated(True)

        
        self.connect(self.listView, SIGNAL("itemSelectionChanged()"), self.ViewSelectionChanged)
        
        # table of significant GO terms
        self.sigTerms = QTreeWidget(self.splitter)
        self.sigTerms.setColumnCount(len(self.DAGcolumns))
        self.sigTerms.setHeaderLabels(self.DAGcolumns)
        self.sigTerms.setSortingEnabled(True)
        self.sigTerms.setSelectionMode(QAbstractItemView.ExtendedSelection)
        
        self.connect(self.sigTerms, SIGNAL("itemSelectionChanged()"), self.TableSelectionChanged)
        self.splitter.show()

        self.sigTableTermsSorted = []
        self.graph = {}
        
        self.loadedAnnotationCode = "---"
        
        self.inputTab.layout().addStretch(1)
        self.filterTab.layout().addStretch(1)
        self.selectTab.layout().addStretch(1)
        
        self.resize(1000, 800)

        self.clusterDataset = None
        self.referenceDataset = None
        self.ontology = None
        self.annotations = None
        self.treeStructRootKey = None
        self.probFunctions = [obiProb.Binomial(), obiProb.Hypergeometric()]
        self.selectedTerms = []
        
        self.connect(self, SIGNAL("widgetStateChanged(QString, int, QString)"), self.onStateChanged)
        
    def UpdateOrganismComboBox(self):
        try:
            if self.annotationCodes and len(self.annotationCodes) > self.annotationIndex:
                currAnnotationCode = self.annotationCodes[self.annotationIndex]
            else:
                currAnnotationCode = None
            self.progressBarInit()
            with orngServerFiles.DownloadProgress.setredirect(self.progressBarSet):
                self.annotationFiles = listAvailable()
            self.progressBarFinished()
            self.annotationCodes = sorted(self.annotationFiles.keys())
            self.annotationComboBox.clear()
            self.annotationComboBox.addItems(self.annotationCodes)
            self.annotationComboBox.setCurrentIndex(self.annotationIndex)
        finally:
            self.signalManager.freeze(self).pop()
            
    def UpdateGeneMatcher(self):
        dialog = GeneMatcherDialog(self, defaults=self.geneMatcherSettings, modal=True)
        if dialog.exec_():
            self.geneMatcherSettings = [getattr(dialog, item[0]) for item in dialog.items]
            if self.annotations:
                self.SetGeneMatcher()
                if self.clusterDataset:
                    self.Update()
                
    def Update(self):
        if self.clusterDataset:
            pb = OWGUI.ProgressBar(self, 100)
            self.Load(pb=pb)
            self.FilterUnknownGenes()
            graph = self.Enrichment(pb=pb)
            self.SetGraph(graph)

    def UpdateGOAndAnnotation(self, tags=[]):
        from .OWUpdateGenomicsDatabases import OWUpdateGenomicsDatabases
        w = OWUpdateGenomicsDatabases(parent = self, searchString=" ".join(tags))
        w.setModal(True)
        w.show()
        self.UpdateAnnotationComboBox()

    def UpdateAnnotationComboBox(self):
        if self.annotationCodes:
            curr = self.annotationCodes[min(self.annotationIndex, len(self.annotationCodes)-1)]
        else:
            curr = None
        self.annotationFiles = listAvailable()
        self.annotationCodes = self.annotationFiles.keys()
        index = curr and self.annotationCodes.index(curr) or 0
        self.annotationComboBox.clear()
        self.annotationComboBox.addItems(self.annotationCodes)
        self.annotationComboBox.setCurrentIndex(index)
        if not self.annotationCodes:
            self.error(0, "No downloaded annotations!!\nClick the update button and update annotationa for at least one organism!")
        else:
            self.error(0)

    def SetGenesComboBox(self):
        self.candidateGeneAttrs = self.clusterDataset.domain.variables + self.clusterDataset.domain.getmetas().values()
        self.candidateGeneAttrs = filter(lambda v: v.varType in [orange.VarTypes.String,
                                                                 orange.VarTypes.Other,
                                                                 orange.VarTypes.Discrete], 
                                         self.candidateGeneAttrs)
        self.geneAttrIndexCombo.clear()
        self.geneAttrIndexCombo.addItems([a.name for a in  self.candidateGeneAttrs])

    def FindBestGeneAttrAndOrganism(self):
        if self.autoFindBestOrg:  
            organismGenes = dict([(o,set(go.getCachedGeneNames(o))) for o in self.annotationCodes])
        else:
            currCode = self.annotationCodes[min(self.annotationIndex, len(self.annotationCodes)-1)]
            filename = p_join(dataDir, self.annotationFiles[currCode])
            try:
                f = tarfile.open(filename)
                info = [info for info in f.getmembers() if info.name.startswith("gene_names")].pop()
                geneNames = cPickle.loads(f.extractfile(info).read().replace("\r\n", "\n"))
            except Exception, ex:
                geneNames = cPickle.loads(open(p_join(filename, "gene_names.pickle")).read().replace("\r\n", "\n"))
            organismGenes = {currCode: set(geneNames)}
        candidateGeneAttrs = self.clusterDataset.domain.attributes + self.clusterDataset.domain.getmetas().values()
        candidateGeneAttrs = filter(lambda v: v.varType in [orange.VarTypes.String, 
                                                            orange.VarTypes.Other, 
                                                            orange.VarTypes.Discrete], 
                                    candidateGeneAttrs)
        attrNames = [v.name for v in self.clusterDataset.domain.variables]
        cn = {}
        for attr in candidateGeneAttrs:
            vals = [str(e[attr]) for e in self.clusterDataset]
            if any("," in val for val in vals):
                vals = reduce(list.__add__, (val.split(",") for val in vals))
            for organism, s in organismGenes.items():
                l = filter(lambda a: a in s, vals)
                cn[(attr,organism)] = len(set(l))
        for organism, s in organismGenes.items():
            l = filter(lambda a: a in s, attrNames)
            cn[("_var_names_", organism)] = len(set(l))
            
        cn = cn.items()
        cn.sort(lambda a,b:-cmp(a[1],b[1]))
        ((bestAttr, organism), count) = cn[0]
        if bestAttr=="_var_names_" and count<=len(attrNames)/10.0 or \
           bestAttr!="_var_names_" and count<=len(self.clusterDataset)/10.0:
            return
        
        self.annotationIndex = self.annotationCodes.index(organism)
        if bestAttr=="_var_names_":
            self.useAttrNames = True
            self.geneAttrIndex = 0
        else:
            self.useAttrNames = False
            self.geneAttrIndex = candidateGeneAttrs.index(bestAttr)
    
    def SetClusterDataset(self, data=None):
        if not self.annotationCodes:
            QTimer.singleShot(200, lambda: self.SetClusterDataset(data))
            return
        self.closeContext()
        self.clusterDataset = data
        self.infoLabel.setText("\n")
        if data:
            self.SetGenesComboBox()
            try:
                taxid = data_hints.get_hint(data, "taxid", "")
                code = obiGO.from_taxid(taxid)
                filename = "gene_association.%s.tar.gz" % code
                if filename in self.annotationFiles.values():
                    self.annotationIndex = [i for i, name in enumerate(self.annotationCodes) \
                                            if self.annotationFiles[name] ==  filename].pop()
            except Exception:
                pass
            self.useAttrNames = data_hints.get_hint(data, "genesinrows", self.useAttrNames)
            self.openContext("", data)
            self.Update()
        else:
            self.infoLabel.setText("No data on input\n")
            self.warning(0)
            self.warning(1)
            self.openContext("", None)
            self.ClearGraph()
            self.send("Selected Examples", None)
            self.send("Unselected Examples", None)
            self.send("Example With Unknown Genes", None)

    def SetReferenceDataset(self, data=None):
        self.referenceDataset=data
        self.referenceRadioBox.buttons[1].setDisabled(not bool(data))
        self.referenceRadioBox.buttons[1].setText("Reference set")
        if self.clusterDataset and self.useReferenceDataset:
            self.useReferenceDataset = 0 if not data else 1
            graph = self.Enrichment()
            self.SetGraph(graph)
        elif self.clusterDataset:
            self.UpdateReferenceSetButton()
            
    def UpdateReferenceSetButton(self):
        allgenes, refgenes = None, None
        if self.referenceDataset:
            try:
                allgenes = self.GenesFromExampleTable(self.referenceDataset)
            except Exception:
                allgenes = []
            refgenes, unknown = self.FilterAnnotatedGenes(allgenes)
        self.referenceRadioBox.buttons[1].setDisabled(not bool(allgenes))
        self.referenceRadioBox.buttons[1].setText("Reference set " + ("(%i genes, %i matched)" % (len(allgenes), len(refgenes)) if allgenes and refgenes else ""))

    def GenesFromExampleTable(self, data):
        if self.useAttrNames:
            genes = [v.name for v in data.domain.variables]
        else:
            attr = self.candidateGeneAttrs[min(self.geneAttrIndex, len(self.candidateGeneAttrs) - 1)]
            genes = [str(ex[attr]) for ex in data if not ex[attr].isSpecial()]
            if any("," in gene for gene in genes):
                self.information(0, "Separators detected in gene names. Assuming multiple genes per example.")
                genes = reduce(list.__add__, (genes.split(",") for genes in genes))
        return genes
        
    def FilterAnnotatedGenes(self, genes):
        matchedgenes = self.annotations.GetGeneNamesTranslator(genes).values()
        return matchedgenes, [gene for gene in genes if gene not in matchedgenes]
        
    def FilterUnknownGenes(self):
        if not self.useAttrNames and self.candidateGeneAttrs:
            geneAttr = self.candidateGeneAttrs[min(self.geneAttrIndex, len(self.candidateGeneAttrs)-1)]
            examples = []
            for ex in self.clusterDataset:
                if not any(self.annotations.genematcher.match(n.strip()) for n in str(ex[geneAttr]).split(",")):
                    examples.append(ex)

            self.send("Example With Unknown Genes", examples and orange.ExampleTable(examples) or None)
        else:
            self.send("Example With Unknown Genes", None)

    def Load(self, pb=None):
        go_files, tax_files = orngServerFiles.listfiles("GO"), orngServerFiles.listfiles("Taxonomy")
        calls = []
        pb, finish = (OWGUI.ProgressBar(self, 0), True) if pb is None else (pb, False)
        count = 0
        if not tax_files:
            calls.append(("Taxonomy", "ncbi_taxnomy.tar.gz"))
            count += 1
        org = self.annotationCodes[min(self.annotationIndex, len(self.annotationCodes)-1)]
        if org != self.loadedAnnotationCode:
            count += 1
            if self.annotationFiles[org] not in go_files:
                calls.append(("GO", self.annotationFiles[org]))
                count += 1
                
        if "gene_ontology_edit.obo.tar.gz" not in go_files:
            calls.append(("GO", "gene_ontology_edit.obo.tar.gz"))
            count += 1
        if not self.ontology:
            count += 1
        pb.iter += count*100
        
        for i, args in enumerate(calls):
            orngServerFiles.localpath_download(*args, **dict(callback=pb.advance))
            
        i = len(calls)
        if not self.ontology:
            self.ontology = obiGO.Ontology(progressCallback=lambda value: pb.advance())
            i+=1
        if org != self.loadedAnnotationCode:
            self.annotations = None
            gc.collect() # Force run garbage collection.
            code = self.annotationFiles[org].split(".")[-3]
            self.annotations = obiGO.Annotations(code, genematcher=obiGene.GMDirect(), progressCallback=lambda value: pb.advance())
            i+=1
            self.loadedAnnotationCode = org
            count = defaultdict(int)
            geneSets = defaultdict(set)

            for anno in self.annotations.annotations:
                count[anno.evidence]+=1
                geneSets[anno.evidence].add(anno.geneName)
            for etype in obiGO.evidenceTypesOrdered:
                self.evidenceCheckBoxDict[etype].setEnabled(bool(count[etype]))
                self.evidenceCheckBoxDict[etype].setText(etype+": %i annots(%i genes)" % (count[etype], len(geneSets[etype])))
        if finish:
            pb.finish()
            
    def SetGeneMatcher(self):
        if self.annotations:
            taxid = self.annotations.taxid
            matchers = []
            for matcher, use in zip([obiGene.GMGO, obiGene.GMKEGG, obiGene.GMNCBI, obiGene.GMAffy], self.geneMatcherSettings):
                if use:
                    try:
                        if taxid == "352472":
                            matchers.extend([matcher(taxid), obiGene.GMDicty(),
                                            [matcher(taxid), obiGene.GMDicty()]])
                            # The reason machers are duplicated is that we want `matcher` or `GMDicty` to
                            # match genes by them self if possible. Only use the joint matcher if they fail.   
                        else:
                            matchers.append(matcher(taxid))
                    except Exception, ex:
                        print ex
            self.annotations.genematcher = obiGene.matcher(matchers)
            self.annotations.genematcher.set_targets(self.annotations.geneNames)
            
    def Enrichment(self, pb=None):
        pb = OWGUI.ProgressBar(self, 100) if pb is None else pb
        if not self.annotations.ontology:
            self.annotations.ontology = self.ontology
            
        if isinstance(self.annotations.genematcher, obiGene.GMDirect):
            self.SetGeneMatcher()
        self.error(1)
        self.warning([0, 1])
        try:    
            if self.useAttrNames:
                clusterGenes = [v.name for v in self.clusterDataset.domain.variables]
                self.information(0)
            else:
                geneAttr = self.candidateGeneAttrs[min(self.geneAttrIndex, len(self.candidateGeneAttrs)-1)]
                clusterGenes = [str(ex[geneAttr]) for ex in self.clusterDataset if not ex[geneAttr].isSpecial()]
                if any("," in gene for gene in clusterGenes):
                    self.information(0, "Separators detected in cluster gene names. Assuming multiple genes per example.")
                    clusterGenes = reduce(list.__add__, (genes.split(",") for genes in clusterGenes))
                else:
                    self.information(0)
        except Exception, ex:
            self.error(1, "Failed to extract gene names from input dataset! %s" % str(ex))
            return {}
        genesCount = len(clusterGenes)
        genesSetCount = len(set(clusterGenes))
        
        self.clusterGenes = clusterGenes = self.annotations.GetGeneNamesTranslator(clusterGenes).values()
        
#        self.clusterGenes = clusterGenes = filter(lambda g: g in self.annotations.aliasMapper or g in self.annotations.additionalAliases, clusterGenes)
        self.infoLabel.setText("%i unique genes on input\n%i (%.1f%%) genes with known annotations" % (genesSetCount, len(clusterGenes), 100.0*len(clusterGenes)/genesSetCount if genesSetCount else 0.0))
        
        referenceGenes = None
        if self.referenceDataset:
            try:
                if self.useAttrNames:
                    referenceGenes = [v.name for v in self.referenceDataset.domain.variables]
                    self.information(1)
                else:
                    referenceGenes = [str(ex[geneAttr]) for ex in self.referenceDataset if not ex[geneAttr].isSpecial()]
                    if any("," in gene for gene in clusterGenes):
                        self.information(1, "Separators detected in reference gene names. Assuming multiple genes per example.")
                        referenceGenes = reduce(list.__add__, (genes.split(",") for genes in referenceGenes))
                    else:
                        self.information(1)

                refc = len(referenceGenes)
#                referenceGenes = filter(lambda g: g in self.annotations.aliasMapper or g in self.annotations.additionalAliases, referenceGenes)
                referenceGenes = self.annotations.GetGeneNamesTranslator(referenceGenes).values()
                self.referenceRadioBox.buttons[1].setText("Reference set (%i genes, %i matched)" % (refc, len(referenceGenes)))
                self.referenceRadioBox.buttons[1].setDisabled(False)
                self.information(2)
            except Exception, er:
                if not self.referenceDataset:
                    self.information(2, "Unable to extract gene names from reference dataset. Using entire genome for reference")
                else:
                    self.referenceRadioBox.buttons[1].setText("Reference set")
                    self.referenceRadioBox.buttons[1].setDisabled(True)
                referenceGenes = self.annotations.geneNames
                self.useReferenceDataset = 0
        else:
            self.useReferenceDataset = 0
        if not self.useReferenceDataset:
            self.information(2)
            self.information(1)
            referenceGenes = self.annotations.geneNames
        self.referenceGenes = referenceGenes
        evidences = []
        for etype in obiGO.evidenceTypesOrdered:
            if getattr(self, "useEvidence"+etype):
                evidences.append(etype)
        aspect = ["P", "C", "F"][self.aspectIndex]
        
        if clusterGenes:
            self.terms = terms = self.annotations.GetEnrichedTerms(clusterGenes, referenceGenes, evidences, aspect=aspect,
                                                                   prob=self.probFunctions[self.probFunc], useFDR=self.useFDR,
                                                                   progressCallback=lambda value:pb.advance() )
        else:
            self.terms = terms = {}
        if not self.terms:
            self.warning(0, "No enriched terms found.")
        else:
            self.warning(0)
            
        pb.finish()
        self.treeStructDict = {}
        ids = self.terms.keys()
        
        self.treeStructRootKey = None
        
        parents = {}
        for id in ids:
            parents[id] = set([term for typeId, term in self.ontology[id].related])
            
        children = {}
        for term in self.terms:
            children[term] = set([id for id in ids if term in parents[id]])
            
        for term in self.terms:
            self.treeStructDict[term] = TreeNode(self.terms[term], children[term])
            if not self.ontology[term].related and not getattr(self.ontology[term], "is_obsolete", False):
                self.treeStructRootKey = term
        return terms
        
    def FilterGraph(self, graph):
        if self.filterByPValue:
            graph = obiGO.filterByPValue(graph, self.maxPValue)
        if self.filterByNumOfInstances:
            graph = dict(filter(lambda (id,(genes, p, rc)):len(genes)>=self.minNumOfInstances, graph.items()))
        return graph

    def FilterAndDisplayGraph(self):
        if self.clusterDataset:
            self.graph = self.FilterGraph(self.originalGraph)
            if self.originalGraph and not self.graph:
                self.warning(1, "All found terms were filtered out.")
            else:
                self.warning(1)
            self.ClearGraph()
            self.DisplayGraph()

    def SetGraph(self, graph=None):
        self.originalGraph = graph
        if graph:
            self.FilterAndDisplayGraph()
        else:
            self.graph = {}
            self.ClearGraph()

    def ClearGraph(self):
        self.listView.clear()
        self.listViewItems=[]
        self.sigTerms.clear()

    def DisplayGraph(self):
        fromParentDict = {}
        self.termListViewItemDict = {}
        self.listViewItems=[]
        enrichment = lambda t:float(len(t[0])) / t[2] * (float(len(self.referenceGenes))/len(self.clusterGenes))
        maxFoldEnrichment = max([enrichment(term) for term in self.graph.values()] or [1])
        def addNode(term, parent, parentDisplayNode):
            if (parent, term) in fromParentDict:
                return
            if term in self.graph:
                displayNode = GOTreeWidgetItem(self.ontology[term], self.graph[term], len(self.clusterGenes), len(self.referenceGenes), maxFoldEnrichment, parentDisplayNode)
                displayNode.goId = term
                self.listViewItems.append(displayNode)
                if term in self.termListViewItemDict:
                    self.termListViewItemDict[term].append(displayNode)
                else:
                    self.termListViewItemDict[term] = [displayNode]
                fromParentDict[(parent, term)] = True
                parent = term
            else:
                displayNode = parentDisplayNode
            
            for c in self.treeStructDict[term].children:
                addNode(c, parent, displayNode)
                
        if self.treeStructDict:
            addNode(self.treeStructRootKey, None, self.listView)

        terms = self.graph.items()
        terms = sorted(terms, key=lambda item: item[1][1])
        self.sigTableTermsSorted = [t[0] for t in terms]
        
        self.sigTerms.clear()
        for i, (id, (genes, p_value, refCount)) in enumerate(terms):
            text = [self.ontology[id].name, str(len(genes)), str(refCount), "%.4f" % p_value, " ,".join(genes), "%.2f" % enrichment((genes, p_value, refCount))]
            item = GOTreeWidgetItem(self.ontology[id], (genes, p_value, refCount), len(self.clusterGenes),
                                    len(self.referenceGenes), maxFoldEnrichment, self.sigTerms)
            item.goId = id
                
        self.listView.expandAll()
        for i in range(4):
            self.listView.resizeColumnToContents(i)
            self.sigTerms.resizeColumnToContents(i)
        self.sigTerms.resizeColumnToContents(5)
        width = min(self.listView.columnWidth(0), 350)
        self.listView.setColumnWidth(0, width)
        self.sigTerms.setColumnWidth(0, width)
        
    def ViewSelectionChanged(self):
        if self.selectionChanging:
            return
        
        self.selectionChanging = 1
        self.selectedTerms = []
        selected = self.listView.selectedItems()
        self.selectedTerms = list(set([lvi.term.id for lvi in selected]))
        self.ExampleSelection()
        self.selectionChanging = 0
        
        
    def TableSelectionChanged(self):
        if self.selectionChanging:
            return
        
        self.selectionChanging = 1
        self.selectedTerms = []
        selectedIds = set([self.sigTerms.itemFromIndex(index).goId for index in self.sigTerms.selectedIndexes()])
        
        for i in range(self.sigTerms.topLevelItemCount()):
            item = self.sigTerms.topLevelItem(i)
            selected = item.goId in selectedIds
            term = item.goId
            
            if selected:
                self.selectedTerms.append(term)
                
            for lvi in self.termListViewItemDict[term]:
                try:
                    lvi.setSelected(selected)
                    if selected: lvi.setExpanded(True)
                except RuntimeError:    ##Underlying C/C++ object deleted (why??)
                    pass
                
        self.ExampleSelection()
        self.selectionChanging = 0
            
    
    def UpdateAddClassButton(self):
        self.addClassCB.setEnabled(self.selectionDisjoint == 1)
        
    def ExampleSelection(self):
        selectedExamples = []
        unselectedExamples = []
        selectedGenes = []
        
        #change by Marko. don't do anything if there is no3 dataset 
        if not self.clusterDataset:
            return
        
        selectedGenes = reduce(set.union, [v[0] for id, v in self.graph.items() if id in self.selectedTerms], set())
        evidences = []
        for etype in obiGO.evidenceTypesOrdered:
            if getattr(self, "useEvidence"+etype):
                evidences.append(etype)
        allTerms = self.annotations.GetAnnotatedTerms(selectedGenes, 
                          directAnnotationOnly=self.selectionDirectAnnotation, 
                          evidenceCodes=evidences)
            
        if self.selectionDisjoint:
            count = defaultdict(int)
            for term in self.selectedTerms:
                for g in allTerms.get(term, []):
                    count[g]+=1
            ccount = 1 if self.selectionDisjoint==1 else len(self.selectedTerms)
            selectedGenes = [gene for gene, c in count.items() if c==ccount and gene in selectedGenes]
        else:
            selectedGenes = reduce(set.union, [allTerms.get(term, []) for term in self.selectedTerms], set())

        if self.useAttrNames:
            vars = [self.clusterDataset.domain[gene] for gene in set(selectedGenes)]
            newDomain = orange.Domain(vars, self.clusterDataset.domain.classVar)
            newdata = orange.ExampleTable(newDomain, self.clusterDataset)
            self.send("Selected Examples", newdata)
            self.send("Unselected Examples", None)
        elif self.candidateGeneAttrs:
            geneAttr = self.candidateGeneAttrs[min(self.geneAttrIndex, len(self.candidateGeneAttrs)-1)]
            if self.selectionDisjoint == 1:
                goVar = orange.EnumVariable("GO Term", values=list(self.selectedTerms))
                newDomain = orange.Domain(self.clusterDataset.domain.variables, goVar)
                newDomain.addmetas(self.clusterDataset.domain.getmetas())
#            else:
#                goVar = orange.StringVariable("GO Terms")
#                newDomain = orange.Domain(self.clusterDataset.domain)
#                newDomain.addmeta(orange.newmetaid(), goVar)
            
            
            for ex in self.clusterDataset:
                if not ex[geneAttr].isSpecial() and any(gene in selectedGenes for gene in str(ex[geneAttr]).split(",")):
                    if self.selectionDisjoint == 1 and self.selectionAddTermAsClass:
                        terms = filter(lambda term: any(gene in self.graph[term][0] for gene in str(ex[geneAttr]).split(",")) , self.selectedTerms)
                        term = sorted(terms)[0]
                        ex =  orange.Example(newDomain, ex)
                        ex[goVar] = goVar(term)
#                        ex.setclass(newClass(term))
                    selectedExamples.append(ex)
                else:
                    unselectedExamples.append(ex)
                    
            if selectedExamples:
                selectedExamples = orange.ExampleTable(selectedExamples)
            else:
                selectedExamples = None
                
            if unselectedExamples:
                unselectedExamples = orange.ExampleTable(unselectedExamples)
            else:
                unselectedExamples = None
            
            self.send("Selected Examples", selectedExamples)
            self.send("Unselected Examples", unselectedExamples)

    def ShowInfo(self):
        dialog = QDialog(self)
        dialog.setModal(False)
        dialog.setLayout(QVBoxLayout())
        label = QLabel(dialog)
        label.setText("Ontology:\n"+self.ontology.header if self.ontology else "Ontology not loaded!")
        dialog.layout().addWidget(label)

        label = QLabel(dialog)
        label.setText("Annotations:\n"+self.annotations.header.replace("!", "") if self.annotations else "Annotations not loaded!")
        dialog.layout().addWidget(label)
        dialog.show()
        
    def sendReport(self):
        self.reportSettings("Settings", [("Organism", self.annotationCodes[min(self.annotationIndex, len(self.annotationCodes) - 1)]),
                                         ("Significance test", ("Binomial" if self.probFunc == 0 else "Hypergeometric") + (" with FDR" if self.useFDR else ""))])
        self.reportSettings("Filter", ([("Min cluster size", self.minNumOfInstances)] if self.filterByNumOfInstances else []) + \
                                      ([("Max p-value", self.maxPValue)] if self.filterByPValue else []))

        def treeDepth(item):
            return 1 + max([treeDepth(item.child(i)) for i in range(item.childCount())] +[0])
        
        def printTree(item, level, treeDepth):
            text = '<tr>' + '<td width=16px></td>' * level
            text += '<td colspan="%i">%s: %s</td>' % (treeDepth - level, item.term.id, item.term.name)
            text += ''.join('<td>%s</td>' % item.text(i) for i in range(1, 4) + [5]) + '</tr>\n'
            for i in range(item.childCount()):
                text += printTree(item.child(i), level + 1, treeDepth)
            return text
        
        treeDepth = max([treeDepth(self.listView.topLevelItem(i)) for i in range(self.listView.topLevelItemCount())] + [0])
        
        tableText = '<table>\n<tr>' + ''.join('<th>%s</th>' % s for s in ["Term:", "List:", "Reference:", "P-value:", "Enrichment:"]) + '</tr>'
        
        treeText = '<table>\n' +  '<th colspan="%i">%s</th>' % (treeDepth, "Term:") 
        treeText += ''.join('<th>%s</th>' % s for s in ["List:", "Reference:", "P-value:", "Enrichment:"]) + '</tr>'
        
        for index in range(self.sigTerms.topLevelItemCount()):
            item = self.sigTerms.topLevelItem(index)
            tableText += printTree(item, 0, 1) 
        tableText += '</table>' 
        
        for index in range(self.listView.topLevelItemCount()):
            item = self.listView.topLevelItem(index)
            treeText += printTree(item, 0, treeDepth)
        
        self.reportSection("Enriched Terms")
        self.reportRaw(tableText)
        
        self.reportSection("Enriched Terms in the Ontology Tree")
        self.reportRaw(treeText)
        
    def onStateChanged(self, stateType, id, text):
        if stateType == "Warning":
            self.listView._userMessage = text
            self.listView.viewport().update()
            
    def onDeleteWidget(self):
        """ Called before the widget is removed from the canvas.
        """
        self.annotations = None
        self.ontology = None
        gc.collect() # Force collection
        

class GOTreeWidgetItem(QTreeWidgetItem):
    def __init__(self, term, enrichmentResult, nClusterGenes, nRefGenes, maxFoldEnrichment, parent):
        QTreeWidgetItem.__init__(self, parent)
        self.term = term
        self.enrichmentResult = enrichmentResult
        self.nClusterGenes = nClusterGenes
        self.nRefGenes = nRefGenes
        self.maxFoldEnrichment = maxFoldEnrichment
        self.enrichment = enrichment = lambda t:float(len(t[0])) / t[2] * (float(nRefGenes)/nClusterGenes)
        self.setText(0, term.name)
        fmt = "%" + str(-int(math.log(nClusterGenes))) + "i (%.2f%%)"
        self.setText(1, fmt % (len(enrichmentResult[0]), 100.0*len(self.enrichmentResult[0])/nClusterGenes))
        fmt = "%" + str(-int(math.log(nRefGenes))) + "i (%.2f%%)"
        self.setText(2, fmt % (enrichmentResult[2], 100.0*enrichmentResult[2]/nRefGenes))
        self.setText(3, "%.4f" % enrichmentResult[1])
        self.setText(4, ", ".join(enrichmentResult[0]))
        self.setText(5, "%.2f" % (enrichment(enrichmentResult)))
        self.setToolTip(0, "<p>" + term.__repr__()[6:].strip().replace("\n", "<br>"))
        self.sortByData = [term.name, len(self.enrichmentResult[0]), enrichmentResult[2], enrichmentResult[1], ", ".join(enrichmentResult[0]), enrichment(enrichmentResult)]

    def data(self, col, role):
        if role == Qt.UserRole:
            return QVariant(self.enrichment(self.enrichmentResult) / self.maxFoldEnrichment)
        else:
            return QTreeWidgetItem.data(self, col, role)

    def __lt__(self, other):
        col = self.treeWidget().sortColumn()
        return self.sortByData[col] < other.sortByData[col]
    
class EnrichmentColumnItemDelegate(QItemDelegate):
    def paint(self, painter, option, index):
        self.drawBackground(painter, option, index)
        value, ok = index.data(Qt.UserRole).toDouble()
        if ok:
            painter.save()
            painter.setBrush(QBrush(Qt.white, Qt.SolidPattern))
            painter.drawRect(option.rect)
            painter.setBrush(QBrush(Qt.blue, Qt.SolidPattern))
            painter.drawRect(option.rect.x(), option.rect.y(), value*(option.rect.width()-1), option.rect.height()-1)
            painter.restore()
        else:
            QItemDelegate.paint(self, painter, option, index)
        
        
class GeneMatcherDialog(OWWidget):
    items = [("useGO", "Use gene names from Gene Ontology annotations"),
             ("useKEGG", "Use gene names from KEGG Genes database"),
             ("useNCBI", "Use gene names from NCBI Gene info database"),
             ("useAffy", "Use Affymetrix platform reference ids")]
    settingsList = [item[0] for item in items]
    def __init__(self, parent=None, defaults=[True, False, False, False], enabled=[False, True, True, True], **kwargs):
        OWWidget.__init__(self, parent, **kwargs)
        for item, default in zip(self.items, defaults):
            setattr(self, item[0], default)
            
        self.loadSettings()
        for item, enable in zip(self.items, enabled):
            cb = OWGUI.checkBox(self, self, *item)
            cb.setEnabled(enable)
            
        box = OWGUI.widgetBox(self, orientation="horizontal")
        OWGUI.button(box, self, "OK", callback=self.accept)
        OWGUI.button(box, self, "Cancel", callback=self.reject)
        
        
if __name__=="__main__":
    import sys
    app = QApplication(sys.argv)
    w=OWGOEnrichmentAnalysis()
    data = orange.ExampleTable("../../../doc/datasets/brown-selected.tab")
    w.show()
    w.SetClusterDataset(data)
    app.exec_()
    w.saveSettings()
        



import math    

def _lngamma(z):
    x = 0
    x += 0.1659470187408462e-06 / (z + 7)
    x += 0.9934937113930748e-05 / (z + 6)
    x -= 0.1385710331296526 / (z + 5)
    x += 12.50734324009056 / (z + 4)
    x -= 176.6150291498386 / (z + 3)
    x += 771.3234287757674 / (z + 2)
    x -= 1259.139216722289 / (z + 1)
    x += 676.5203681218835 / (z)
    x += 0.9999999999995183
    
    return math.log(x) - 5.58106146679532777 - z + (z - 0.5) * math.log(z + 6.5)
        
class LogBin(object):
    _max = 2
    _lookup = [0.0, 0.0]
    _max_factorial = 1
    def __init__(self, max=1000):
        self._extend(max)

    @staticmethod
    def _extend(max):
        if max <= LogBin._max:
            return
        for i in range(LogBin._max, max):
            if i > 1000: ## an arbitrary cuttof
                LogBin._lookup.append(LogBin._logfactorial(i))
            else:
                LogBin._max_factorial *= i
                LogBin._lookup.append(math.log(LogBin._max_factorial))
        LogBin._max = max

    def _logbin(self, n, k):
        if n >= self._max:
            self._extend(n + 100)
        if k < n and k >= 0:
            return self._lookup[n] - self._lookup[n - k] - self._lookup[k]
        else:
            return 0.0

    @staticmethod
    def _logfactorial(n):
        if (n <= 1):
            return 0.0
        else:
            return _lngamma(n + 1)

class Binomial(LogBin):

    def __call__(self, k, N, m, n):
        p = 1.0 * m / N
        if p == 0.0:
            if k == 0:
                return 1.0
            else:
                return 0.0
        elif p == 1.0:
            if n == k:
                return 1.0
            else:
                return 0.0
        try:
            return min(math.exp(self._logbin(n, k) + k * math.log(p) + (n - k) * math.log(1.0 - p)), 1.0)
        except (OverflowError, ValueError), er:
            print k, N, m, n
            raise
##        return math.exp(self._logbin(n, k) + math.log((p**k) * (1.0 - p)**(n - k)))

    def p_value(self, k, N, m, n):
        subtract = n - k + 1 > k
        result = sum([self.__call__(i, N, m, n) for i in (range(k) if subtract else range(k, n+1))])
        return max(1.0 - result if subtract else result, 0.0)

class Hypergeometric(LogBin):

    def __call__(self, k, N, m, n):
        if k < max(0, n + m - N) or k > min(n, m):
            return 0.0
        try:
            return min(math.exp(self._logbin(m, k) + self._logbin(N - m, n - k) - self._logbin(N, n)), 1.0)
        except (OverflowError, ValueError), er:
            print k, N, m, n
            raise

    def p_value(self, k, N, m, n):
        subtract = n - k + 1 > k
##        result = sum([math.exp(self._logbin(m, i) + self._logbin(N - m, n - i)) for i in (range(k) if subtract else range(k, n+1))])
        result = sum([self.__call__(i, N, m, n) for i in (range(k) if subtract else range(k, n+1))])
##        result /= math.exp(self._logbin(N, n))
        return max(1.0 - result if subtract else result, 0.0)


## to speed-up FDR, calculate ahead sum([1/i for i in range(1, m+1)]), for m in [1,100000]. For higher values of m use an approximation, with error less or equal to 4.99999157277e-006. (sum([1/i for i in range(1, m+1)])  ~ log(m) + 0.5772..., 0.5572 is an Euler-Mascheroni constant) 
c = [1.0]
for m in range(2, 100000):
    c.append( c[-1] + 1.0/m)

def is_sorted(l):
    return all(l[i] <= l[i+1] for i in xrange(len(l)-1))

def FDR(p_values, dependent=False, m=None, ordered=False):
    """
    If the user is sure that pvalues as already sorted nondescendingly
    setting ordered=True will make the computation faster.
    """

    if not ordered:
        ordered = is_sorted(p_values)

    if not ordered:
        joined = [ (v,i) for i,v in enumerate(p_values) ]
        joined.sort()
        p_values = [ p[0] for p in joined ]
        indices = [ p[1] for p in joined ]

    if not m:
        m = len(p_values)
    if m <= 0 or not p_values:
        return []

    if dependent: # correct q for dependent tests
        k = c[m-1] if m <= len(c) else math.log(m) + 0.57721566490153286060651209008240243104215933593992
        m = m * k

    tmp_fdrs = [p*m/(i+1.0) for (i, p) in enumerate(p_values)]
    fdrs = []
    cmin = tmp_fdrs[-1]
    for f in reversed(tmp_fdrs):
        cmin = min(f, cmin)
        fdrs.append( cmin)
    fdrs.reverse()

    if not ordered:
        new = [ None ] * len(fdrs)
        for v,i in zip(fdrs, indices):
            new[i] = v
        fdrs = new

    return fdrs

def Bonferroni(p_values, m=None):
    if not m:
        m = len(p_values)
    if m == 0:
        return []
    m = float(m)
    return [p/m for p in p_values]



"""
=======
obiKEGG
=======
obiKEGG module is an interface to `Kyoto Encyclopedia of Genes and Genomes
<http://www.genome.jp/kegg/>`_ that allows easy access to KEGG pathway
and genes data.

Example ::
    
    >>> genes = KEGGOrganism("human")
    >>> genes
"""

from __future__ import absolute_import

try:
    import Image, ImageDraw, ImageMath
except:
    pass

import cStringIO
import math
import time
import os, sys, tarfile
import re

from cPickle import load, loads, dump
from collections import defaultdict

import xml.dom.minidom as minidom

from functools import partial, wraps
import urllib2
import cPickle

from Orange.orng import orngEnviron, orngServerFiles

from . import obiData, obiProb, obiTaxonomy

DEFAULT_DATABASE_PATH = orngServerFiles.localpath("KEGG")
KEGG_FTP_PATH = "ftp://ftp.genome.jp/pub/kegg/"

_join = lambda *args: os.path.join(DEFAULT_DATABASE_PATH, *args)

def special_tags(domain, filename):
    return dict([tuple(tag.split(":", 1)) for tag in orngServerFiles.info(domain, filename)["tags"] \
                 if tag.startswith("#") and ":" in tag])
def loads(domain, filename, version=None):
    """ A function decorator for a function that loads data from server files. 
    Makes sure the file is downloaded and of the right version
    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                ## First try downloading the required server file  
                file = filename if isinstance(filename, basestring) else filename(*args, **kwargs)
                orngServerFiles.localpath_download(domain, file)
                if version is not None and special_tags(domain, file).get("#version", str(version)) < str(version):
                    orngServerFiles.update(domain, file) 
            except Exception, ex:
                print ex, args, kwargs
            return func(*args, **kwargs)
        return wrapper
    return decorator
            
def deprecated(first_arg, text=None):
    import warnings
    def wrapper(*args, **kwargs):
        warnings.warn_explicit(("Call to deprecated function %s. " % first_arg.__name__) + (text or ""), 
                               category=DeprecationWarning,
                               filename=first_arg.func_code.co_filename,
                               lineno=first_arg.func_code.co_firstlineno + 1
                      )
#        warnings.warn(("Call to deprecated function %s." % first_arg.__name__) + (text or ""), 
#                               category=DeprecationWarning,;
#                               stacklevel =  -1 #-2 if text is None else -3
#                        )
        return first_arg(*args, **kwargs)
    if isinstance(first_arg, basestring): #deprecated called with a string as first argument
        return partial(deprecated, text=first_arg)
    else:
        return wraps(first_arg)(wrapper)
        
def cached(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        sig = args + tuple(sorted(kwargs.items()))
        if sig in wrapper:
            pass
    return wrapper
            
def cached_property(func):
    @property
    @wraps(func)
    def wrapper_func(self):
        cache_name = "_" + func.__name__
        if not hasattr(self, cache_name):
            setattr(self, cache_name, func(self))
        return getattr(self, cache_name)
    return wrapper_func

def cached_method(func, cache_name="_cached_method_cache", store=None):
    def wrapper(self, *args, **kwargs):
        sig = (func.__name__,) + args + tuple(sorted(kwargs.items()))
        if not hasattr(self, cache_name):
            setattr(self, cache_name, store() if store is not None else {})
        if sig not in getattr(self, cache_name):
            getattr(self, cache_name)[sig] = func(self, *args, **kwargs)
        return getattr(self, cache_name)[sig]
    return wrapper

class pickle_cache(dict):
    def __init__(self, filename, sync=True):
        dict.__init__(self)
        self.filename = filename
        if os.path.exists(self.filename):
            try:
                self.update(cPickle.load(open(self.filename, "rb")))
            except Exception, ex:
                print self.filename, "loading failed", ex
                
    def __setitem__(self, key, value):
        super(pickle_cache, self).__setitem__(key, value)
        self.sync()
        
    def __delitem__(self, key):
        super(pickle_cache, self).__delitem__(key)
        self.sync()
        
    def sync(self):
#        print "sync", self, self.filename
        cPickle.dump(dict(self.items()), open(self.filename, "wb"), 2)
        
def persistent_cached_method(func):
    return cached_method(func, cache_name="_persistent_cached_method_cache")

def persistent_cached_class(cls, store=pickle_cache):
    def cache_loader(self):
        if not hasattr(self, "_cache_filename"):
            setattr(self, "_cache_filename", getattr(self, "filename") + ".cache")
#        print "creating cache", self._cache_filename 
        try:
#            print os.stat(self.filename).st_mtime, os.stat(self._cache_filename).st_mtime
            if os.stat(self.filename).st_mtime > os.stat(self._cache_filename).st_mtime:
#                print "cache obsolete"
                os.remove(self._cache_filename)
        except Exception, ex:
            pass
        return store(self._cache_filename)
    cls._persistent_cached_method_cache = cached_property(cache_loader)
    return cls

class cache_manager(object):
    """ takes an fuction that returns a dict like object (cache)
    """
    def __init__(self, opener):
        self.opener = opener
        self.name = opener.__name__
        
    def __call__(self, *args, **kwargs):
        return self.opener(self, *args, **kwargs)
            
    def __get__(self, instance, type_=None):
        """ Return an cache store from instance """
        ## TODO: check if instance is a type
        if getattr(instance, self.name):
            return getattr(instance, self.name)
        return self.opener.__get__(instance, type)
    
    def cached_method(manager, wrapped):
        def wrapper(self, *args, **kwargs):
            sig = (wrapped.__name__,) + args + tuple(sorted(kwargs.items()))
            cache = getattr(self, manager.__name__)
            if sig not in cache:
                cache[sig] = func(self, *args, **kwargs)
            return cache[sig]

def proxy_dict_decorator(cls, proxy_name):
    for name in ["__len__", "__contains__", "__getitem__", "__iter__", "get", "has_key", "items", 
                 "iteritems", "iterkeys", "itervalues", "keys", "values", "update"]:
        def proxy(func):
            def f(self, *args, **kwargs):
                return func(getattr(self, proxy_name), *args, **kwargs)
            return f
        setattr(cls, name, proxy(getattr(dict, name)))
    return cls

def _create_path_for_file(target): 
    try:
        os.makedirs(os.path.dirname(target))
    except OSError:
        pass

def downloader(func):
    def wrapper(*args, **kwargs):
        import shutil
        def download(src, dst):
            if isinstance(dst, basestring):
                _create_path_for_file(dst)
                dst = open(dst, "wb")
            shutil.copyfileobj(src, dst)
        jobs = func(*args, **kwargs)
        for src, dst in [jobs] if type(jobs) == tuple else jobs:
            download(src, dst)
    return wrapper

from Orange.orng.orngMisc import with_gc_disabled

SLINE, MLINE, LINELIST = range(1, 4)

class _field_mutators(object):
    @classmethod
    def deindent(cls, text, field):
        pad = len(text) - len(text[len(field):].lstrip())
        return "\n".join([line[pad:] for line in text.splitlines()])
    
    @classmethod
    def to_list(cls, text, field):
        return [line for line in text.splitlines() if line.strip()]
    
    @classmethod
    def to_id_list(cls, text, field):
        return [tuple(line.split(": ", 1)) for line in cls.to_list(cls.deindent(text, field), field)]
    
    @classmethod
    def to_dict(cls, text, field):
        text = cls.deindent(text, field).replace("\n ", " ")
        lines = [t.split(": ",1) for t in text.split("\n")]
        return dict([(key, [t.strip() for t in value.split(" ") if t]) for key, value in lines])
    
    @classmethod
    def to_ids(cls, text, field):
        lines = [line.split(" ", 2)[:2] for line in cls.deindent(text, field).split("\n") if ":" in line.split(" ", 1)[0]]
        return [db.lower() + id for db, id in lines] 
#        ids = reduce(lambda col, line:1, text.split("\n"), [])
        return ids
    default = deindent
#    entry = classmethod(lambda cls, text, field: cls.deindent(text, field).split()[0])
    orthology = to_id_list
    pathway = classmethod(lambda cls, text, field: [t[1].split()[0] for t in cls.to_id_list(text, field)])
    codon_usage = classmethod(lambda cls, text, field: text.replace(field, " " * len(field)))
    structure = to_id_list
    motif = to_dict
    pathway = to_ids
    orthology = to_ids
    
    @classmethod
    def genes(cls, text, field):
        genes = [line.split(": ", 1) for line in cls.deindent(text, field).replace("\n    ", "").splitlines()]
        return dict([(org.lower(), [gene.split("(")[0] for gene in genes.split(" ")]) for org, genes in genes])
    
    @classmethod
    def dblinks(cls, text, field):
        links = [line.split(": ", 1) for line in cls.deindent(text, field).replace("\n   ", "").splitlines()]
        return dict([(db.lower(), [name for name in links.split(" ")]) for db, links in links])
    
    dblinks = to_dict

def entry_decorate(cls):
    reserved_map = {"class": "class_"}
    for field in cls.FIELDS:
        mutator = getattr(_field_mutators, field.lower(), _field_mutators.default)
        def construct(field, mutator):
            def get(self):
                text = self.get(field)
                return mutator(text, field) if text else None
            return get
        if not hasattr(cls, field.lower()):
            setattr(cls, reserved_map.get(field.lower(), field.lower()), property(construct(field, mutator)))
        else:
            import warnings
            warnings.warn(str(cls)+ "already contains " + field.lower())
    return cls

def borg_class(cls):
    borg__init__ = cls.__init__
    cls._borg_instance = None
    def __init__(self, *args, **kwargs):
        if not cls._borg_instance:
            borg__init__(self, *args, **kwargs)
            cls._borg_instance = args + tuple(sorted(kwargs.items())), self
        elif cls._borg_instance[0] == args + tuple(sorted(kwargs.items())):
            self.__dict__ = cls._borg_instance[1].__dict__
        else:
            print "Warning. Atempting to create an individual instance of '%s'.\nWe are borg. You will be assimilated. Resistance is futile." % cls.__name__
            self.__dict__ = cls._borg_instance[1].__dict__
    cls.__init__ = __init__
    return cls

def defaultpath(func):
#    @wraps
    def wrapper(*args, **kwargs):
        if "path" not in kwargs:
            kwargs["path"] = DEFAULT_DATABASE_PATH
        return func(*args, **kwargs)
    return wrapper
             
        
class KEGGDBEntry(object):
    FIELDS = []
    MULTIPLE_FIELDS = []
    def __init__(self, entrytext, index=None):
        self.entrytext = entrytext
        if not self.FIELDS:
            self.FIELDS = set([line.split()[0] for line in entrytext.split("\n") if line.strip() and not line.startswith(" ")])
        self._index(index)
        
    def _indices(self, field):
        text = self.entrytext
        fieldlen = len(field)
        return reduce(lambda indices,s: indices.append(text.find(field, indices[-1] + fieldlen)) or indices, range(text.count(field)), [-fieldlen])[1:]
         
    def _index(self, index=None):
        if index is not None:
            index, self.partitions = index
            self.index = dict(zip(self.FIELDS, index))
            return 
        index = dict([(field, self._indices(field)) for field in self.FIELDS])
        sorted_indices = sorted(reduce(set.union, index.values(), set())) + [-1]
        self.partitions = sorted_indices #[(sorted_indices[i], sorted_indices[i+1]) for i in range(len(sorted_indices) - 1)]
        self.index = dict([(key, [sorted_indices.index(ind) for ind in indices]) for key, indices in index.items()])
        
    def section_iter(self, field):
        for ind in self.index[field]:
            yield self.entrytext[self.partitions[ind]: self.partitions[ind+1]]
        
    def get(self, key, default=None):
        if key not in self.FIELDS:
            return default
        sections = [section for section in self.section_iter(key)]
        if key in self.MULTIPLE_FIELDS:
            return sections
        elif sections:
            return sections[0]
        else: 
            return default
        
    def entry_key(self):
        return self.entry.split()[0]
        
class KEGGDataBase(object):
    Entry = KEGGDBEntry
    VERSION = "v1.0"
    def __init__(self, file=None):
        self.load(file)
        
    @with_gc_disabled
    def load(self, file=None):
        if file is None:
            file = self.FILENAME % {"path":DEFAULT_DATABASE_PATH}
        self.filename = file
        data = open(file, "rb").read().split("///\n")
        file_timestamp = os.stat(self.filename).st_mtime
#        print file_timestamp
        build_index = False
#        if os.path.exists(self.filename + ".index"):
        try:
            version, timestamp, index = cPickle.load(open(self.filename + ".index", "rb"))
            assert(version == self.VERSION and timestamp == file_timestamp)
        except Exception, ex:
#            print "error", ex
            build_index = True
            index = [None] * len(data)
        self.entrys = [self.Entry(entry, ind) for entry, ind in zip(data, index) if entry.strip()]
        if build_index:
#            self.entrys = [self.Entry(entry) for entry in data if entry.strip()]
            index = [([e.index[field] for field in e.FIELDS], e.partitions) for e in self.entrys]
            cPickle.dump((self.VERSION, file_timestamp, index), open(self.filename + ".index", "wb"), cPickle.HIGHEST_PROTOCOL)
        
        self.entry_dict = dict([(entry.entry_key(), entry) for entry in self.entrys])
        
proxy_dict_decorator(KEGGDataBase, "entry_dict")
        
class KEGGGeneEntry(KEGGDBEntry):
    FIELDS = ["ENTRY", "NAME", "DEFINITION","ORTHOLOGY", "PATHWAY", "CLASS", "POSITION",
              "MOTIF", "DBLINKS", "STRUCTURE", "CODON_USAGE", "AASEQ", "NTSEQ"]
    
    def aliases(self):
        return [self.entry_key()] + (self.name.split(",") if self.name else []) + [link[1][0] for link in self.dblinks.items() if self.dblinks]
    
    @property
    def alt_names(self):
        return self.aliases()
        
entry_decorate(KEGGGeneEntry)
    
class KEGGGenes(KEGGDataBase):
    Entry = KEGGGeneEntry
    FILENAME = "genes/organisms/%(org_code)s/%(org_name)s.ent"
    
    @loads("KEGG", lambda self, org: "kegg_genes_%s.tar.gz" % org)
    def load(self, org):
        super(KEGGGenes, self).load(_join(self.filename(org)))
        self.entry_dict = dict([(org + ":" + key, value) for key, value in self.entry_dict.items()])
        self.org_code = org
        
    def gene_aliases(self):
        aliases = {}
        for entry in self.entrys:
            aliases.update(dict.fromkeys(entry.aliases(), self.org_code + ":" + entry.entry_key()))
        return aliases
            
    @classmethod
    def filename(cls, org):
        return cls.FILENAME % {"org_code":org, "org_name":KEGGGenome()[org].name.split(",")[1].strip()}
     
    @classmethod
    @downloader
    def download(cls, org):
#        local_dir = orngServerFiles.localpath("KEGG") if local_dir is None else local_dir
        filename = cls.filename(org)
        return (urllib2.urlopen(KEGG_FTP_PATH + filename), 
                _join(filename))
    
class KEGGGenomeEntry(KEGGDBEntry):
    FIELDS = ["ENTRY", "NAME", "DEFINITION", "ANNOTATION", "TAXONOMY", "DATA_SOURCE",
              "ORIGINAL_DB", "CHROMOSOME", "STATISTICS", "REFERENCE"]
    MULTIPLE_FIELDS = ["REFERENCE"]
    
    def org_code(self):
        if self.name is not None:
            return self.name.split(",")[0]
        else:
            return self.entry.split()[0]
        
    def entry_key(self):
        return self.org_code()
    
entry_decorate(KEGGGenomeEntry)
        
class KEGGGenome(KEGGDataBase):
    Entry = KEGGGenomeEntry
    VERSION = "v1.1"
    TAXID_MAP = {"4932" : "559292"}
    def name(self):
        return super(KEGGGeneome, self).name()
    
    @loads("KEGG", "kegg_genome.tar.gz", version=VERSION)
    def load(self, file=None):
        filename = _join("genes", "genome")
        super(KEGGGenome, self).load(filename)
    
    @classmethod
    def common_organisms(cls):
#        genome = KEGGGenome()
#        id_map = {"562":"511145", "2104":"272634", "5833":"36329", "4896":"284812", "11103":None, "4754":None, "4577":None}
#        return [genome.get(id_map.get(taxid, taxid)).entry_key() for taxid in obiTaxonomy.common_taxids() if id_map.get(taxid, taxid) is not None]
        return ['ath', 'bta', 'cel', 'cre', 'dre', 'ddi', 'dme', 'eco', 'hsa', 'mmu', 'mpn', 'osa',
                'pfa', 'rno', 'sce', 'spo', 'zma', 'xla']
        
    @classmethod
    def essential_organisms(cls):
        genome = KEGGGenome()
        
        return [genome.get(cls.TAXID_MAP.get(taxid, taxid)).entry_key() for taxid in obiTaxonomy.essential_taxids()]
            
    def get(self, key, *args):
        if key in self.TAXID_MAP:
            key = self.TAXID_MAP[key]
        if key not in self:
            keys = self.search(key)
            keys = [k for k in keys if key in self[k].name or key in self[k].taxonomy]
            key = keys.pop(0) if keys else key
        return super(KEGGGenome, self).get(key, *args)
    
    def search(self, string, relevance=False):
        if string in self.TAXID_MAP:
            string = self.TAXID_MAP[string]
            
        def match(entry, string):
            rel = 0
            if string in entry.entrytext:
                weight_f = [(entry.definition or "", 2), (entry.name or "", 4), (entry.taxonomy or "", 8), (entry.entry_key(), 16)]
                rel += 1 + sum([w for text, w in weight_f if string in text]) + (16 if entry.entry_key() in self.common_organisms() else 0)
            return rel
        matched = sorted(zip([match(entry, string) for entry in self.entrys], self.entrys), reverse=True)
        return [(entry.entry_key(), rel) if relevance else entry.entry_key() for rel, entry in matched if rel]
            
    @classmethod
    @downloader
    def download(cls, file=None):
        return (urllib2.urlopen(KEGG_FTP_PATH + "/genes/genome"), 
                _join("genes", "genome") if file is None else file)
        
borg_class(KEGGGenome)
        
class KEGGCompoundEntry(KEGGDBEntry):
    FIELDS = ["ENTRY", "NAME", "FORMULA", "MASS", "REMARK", "REACTION", "PATHWAY",
              "ENZYME", "DBLINKS", "ATOM", "BOND"]
    def entry_key(self):
        return "cpd:" + super(KEGGCompoundEntry, self).entry_key()
    
entry_decorate(KEGGCompoundEntry)
    
class KEGGCompounds(KEGGDataBase):
    Entry = KEGGCompoundEntry
    FILENAME = "%(path)s/ligand/compound/compound"
    
    @loads("KEGG", "kegg_ligand.tar.gz")
    def load(self, file=None):
        super(KEGGCompounds, self).load(file)
        
    @classmethod
    @downloader
    def download(cls, file=None):
        return (urllib2.urlopen(KEGG_FTP_PATH + "/ligand/compound/compound"),
                _join("ligand", "compound", "compound") if file is None else file)
    
borg_class(KEGGCompounds)
    
class KEGGEnzymeEntry(KEGGDBEntry):
    FIELDS = ["ENTRY", "NAME", "CLASS", "SYSNAME", "REACTION", "ALL_REAC", "SUBSTRATE",
              "PRODUCT", "COFACTOR", "COMMENT", "REFERENCE", "PATHWAY", "ORTHOLOGY", 
              "GENES", "STRUCTURES", "DBLINKS"]
    MULTIPLE_FIELDS = ["REFERENCE"]
    def entry_key(self):
        return "EC:" + self.entry.split()[1]
    
entry_decorate(KEGGEnzymeEntry)

class KEGGEnzymes(KEGGDataBase):
    Entry = KEGGEnzymeEntry
    FILENAME = "%(path)s/ligand/enzyme/enzyme"

    @loads("KEGG", "kegg_ligand.tar.gz")
    def load(self, file=None):
        super(KEGGEnzymes, self).load(file)
    
    @classmethod
    @downloader
    def download(cls, file=None):
        return (urllib2.urlopen(KEGG_FTP_PATH + "ligand/enzyme/enzyme"),
                _join("ligand", "enzyme", "enzyme"))
    
borg_class(KEGGEnzymes)

class KEGGReactionEntry(KEGGDBEntry):
    FIELDS = ["ENTRY", "NAME", "DEFINITION", "EQUATION", "COMMENT", "RPAIR", "PATHWAY", 
              "ENZYME", "ORTHOLOGY"]
    def entry_key(self):
        return "rn:" + super(KEGGReactionEntry, self).entry_key()
    
entry_decorate(KEGGReactionEntry)

class KEGGReactions(KEGGDataBase):
    Entry = KEGGReactionEntry
    FILENAME = "%(path)s/ligand/reaction/reaction"
    
    @loads("KEGG", "kegg_ligand.tar.gz")
    def load(self, file=None):
        super(KEGGReactions, self).load(file)
        
    @classmethod
    @downloader
    def download(cls, file=None):
        return (urllib2.urlopen(KEGG_FTP_PATH + "ligand/reaction/reaction"),
                _join("ligand", "reaction", "reaction"))
        
borg_class(KEGGReactions)

class KEGGKOEntry(KEGGDBEntry):
    FIELDS = ["ENTRY", "NAME", "DEFINITION", "CLASS", "DBLINKS", "GENES"]
    
    def entry_key(self):
        return "ko:" + super(KEGGKOEntry, self).entry_key()
    
entry_decorate(KEGGKOEntry)
    
class KEGGKO(KEGGDataBase):
    Entry = KEGGKOEntry
    FILENAME = "%(path)s/genes/ko"
    
    @loads("KEGG", "kegg_orthology.tar.gz")
    def load(self, file=None):
        super(KEGGKO, self).load(file)
        
    @classmethod
    @downloader
    def download(cls, file=None):
        return (urllib2.urlopen(KEGG_FTP_PATH + "genes/ko"),
                _join("genes", "ko"))
        
borg_class(KEGGKO)

class KEGGBriteEntry(object):
    _search_re = {"ids": re.compile('(?P<ids>\[.*:.*\])'),
                  "title": re.compile(r'(<[Bb]>)?(?P<title>\b[a-zA-Z0-9_/\s,;:.+=\-\[\]{}\(\)]+?)(?(1)</[Bb]>)$'),
                  "links": re.compile('(?P<links><a href=".+?">.*?</a>)')}
    def __init__(self, line, entrys=None):
        self.entrys = [] #entrys if entrys is not None else []
        self.line = line[1:].strip()
        for name, re in self._search_re.items():
            search = re.search(self.line)
            setattr(self, name, search.group(name) if search else None)
#        self.identifiers = self.groups.get("ids")
#        self.title = self.groups. # TODO: parse line (html, kegg identifiers ...)

    def __iter__(self):
        return iter(self.entrys)

class KEGGBrite(KEGGBriteEntry):
    VERSION = "v1.0"
    def __init__(self, brite_id):
        super(KEGGBrite, self).__init__("")
        self.load(brite_id)
        
    @loads("KEGG", "kegg_brite.tar.gz")
    def load(self, brite_id):
        file = self.filename(brite_id)
        lines = open(file, "rb").read().split("\n!\n")[1].splitlines()
        def collect(lines, depth, collection):
            while lines:
                line = lines[0]
                if line.startswith("#"):
                    lines.pop(0)
                elif line.startswith(depth) and len(line.strip()) > 1:
                    collection.append(KEGGBriteEntry(lines.pop(0))) 
                elif line[0] > depth:
                    collect(lines, line[0], collection[-1].entrys)
                elif line[0] < depth:
                    return
                else:
                    lines.pop(0)
                        
        collect([line for line in lines if not line.startswith("#") and len(line) > 1], "A", self.entrys)
    
    @classmethod
    @defaultpath
    def filename(cls, brite_id, path=DEFAULT_DATABASE_PATH):
        if path is None:
            path = "%(path)s"
        if brite_id.startswith("br"):
            return ("%(path)s/brite/br/" + brite_id + ".keg") % dict(path=path)
        elif brite_id.startswith("ko"):
            return ("%(path)s/brite/ko/" + brite_id + ".keg") % dict(path=path)
        else:
            org = brite[:-5]
            return ("%(path)s/brite/organisms/" + org + "/" + brite_id + ".keg") % dict(path=path)
        
    @classmethod
    @downloader
    def download(cls, brite_id):
        filename = cls.filename(brite_id, path="")
        return (urllib2.urlopen(cls.filename(brite_id, path=KEGG_FTP_PATH.rstrip("/"))),
                cls.filename(brite_id))

class KEGGOrganism(object):
    VERSION = "v2.0"
    DOMAIN = "KEGG"
    def __init__(self, org, genematcher=None):
        self.load(org)
        self.genematcher = genematcher
        
#    @loads("KEGG", lambda self, org: "kegg_genes_%s.tar.gz" % self.organism_name_search(org))
    def load(self, org):
        org = self.organism_name_search(org)
        self.org_code = org
    
    @property
#    @deprecated("Use org_code instead")
    def org(self):
        return self.org_code
    
    @cached_property
    def genes(self):
        return KEGGGenes(self.org_code)
        
    @cached_property
    def gene_aliases(self):
        return self.genes.gene_aliases()
    
    def pathways(self, with_ids=None):
        return [pathway for pathway, values in KEGGPathway.list(self.org_code).items() if all([id in values for id in (with_ids or [])])]
    
#    @deprecated("Use KEGGOrganism.pathways instead")
    def list_pathways(self):
        return self.pathways()
    
#    @deprecated
    def get_linked_pathways(self, pathway_id):
        return KEGGPathway(pathway_id).genes()
    
    def enzymes(self, genes=None):
        enzymes = KEGGEnzymes()
        return [enzyme.entry_key() for enzyme in enzymes.itervalues() if enzyme.genes and self.org_code in enzyme.genes]
    
    def get_enriched_pathways(self, genes, reference=None, prob=obiProb.Binomial(), callback=None):
        """Return a dictionary with enriched pathways ids as keys and (list_of_genes, p_value, num_of_reference_genes) tuples as items."""
        allPathways = defaultdict(lambda :[[], 1.0, []])
        from Orange.orng import orngMisc
        milestones = orngMisc.progressBarMilestones(len(genes), 100)
        for i, gene in enumerate(genes):
            pathways = self.pathways([gene])
            for pathway in pathways:
                allPathways[pathway][0].append(gene)
            if callback and i in milestones:
                callback(i*100.0/len(genes))
        reference = set(reference if reference is not None else self.genes.keys())
        for p_id, entry in allPathways.items():
            entry[2].extend(reference.intersection(KEGGPathway(p_id).genes()))
            entry[1] = prob.p_value(len(entry[0]), len(reference), len(entry[2]), len(genes))
        return dict([(pid, (genes, p, len(ref))) for pid, (genes, p, ref) in allPathways.items()])
    
    def get_genes_by_enzyme(self, enzyme):
        enzyme = KEGGEnzymes()[enzyme]
        return enzyme.genes.get(self.org_code, []) if enzyme.genes else []
    
    def get_genes_by_pathway(self, pathway_id):
        return KEGGPathway(pathway_id).genes()
    
    def get_enzymes_by_pathway(self, pathway_id):
        return KEGGPathway(pathway_id).enzymes()
    
    def get_compounds_by_pathway(self, pathway_id):
        return KEGGPathway(pathway_id).compounds()
    
    def get_pathways_by_genes(self, gene_ids):
        gene_ids = set(gene_ids)
        pathways = [self.genes[id].pathway for id in gene_ids if self.genes[id].pathway]
        pathways = reduce(set.union, pathways, set())
        return [id for id in pathways if gene_ids.issubset(KEGGPathway(id).genes())] 
    
    def get_pathways_by_enzymes(self, enzyme_ids):
        enzyme_ids = set(enzyme_ids)
        pathways = [KEGGEnzymes()[id].pathway for id in enzyme_ids]
        pathwats = reduce(set.union, pathways, set())
        return [id for id in pathways if enzyme_ids.issubset(KEGGPathway(id).enzymes())]
    
    def get_pathways_by_compounds(self, compound_ids):
        compound_ids = set(compound_ids)
        pathways = [KEGGCompounds()[id].pathway for id in compound_ids]
        pathwats = reduce(set.union, pathways, set())
        return [id for id in pathways if compound_ids.issubset(KEGGPathway(id).compounds())]
    
    def get_enzymes_by_compound(self, compound_id):
        return KEGGCompound()[compound_id].enzyme
    
    def get_enzymes_by_gene(self, gene_id):
        return self.genes[gene_id].enzymes
    
    def get_compounds_by_enzyme(self, enzyme_id):
        return self._enzymes_to_compounds.get(enzyme_id)
    
    def get_unique_gene_ids(self, genes, caseSensitive=True):
        """Return a tuple with three elements. The first is a dictionary mapping from unique gene
        ids to gene names in genes, the second is a list of conflicting gene names and the third is a list
        of unknown genes.
        """
        unique, conflicting, unknown = {}, [], []
        for gene in genes:
            names = self.genematcher.match(gene)
            if len(names) == 1:
                unique[names[0]] = gene
            elif len(names) == 0:
                unknown.append(gene)
            else:
                conflicting.append(gene)
        return unique, conflicting, unknown
    
    @property
    @cached_method
    def _enzymes_to_compounds(self):
        dd = {}
        for val in KEGGCompounds().values():
            dd.update(dict.fromkeys(val.enzymes, val.entry_key()))
        return dd
    
    def _set_genematcher(self, genematcher):
        setattr(self, "_genematcher", genematcher)
        
    def _get_genematcher(self):
        if getattr(self, "_genematcher", None) == None:
            from . import obiGene
            if self.org_code == "ddi":
                self._genematcher = obiGene.matcher([obiGene.GMKEGG(self.org_code), obiGene.GMDicty(),
                                                     [obiGene.GMKEGG(self.org_code), obiGene.GMDicty()]])
            else:
                self._genematcher = obiGene.matcher([obiGene.GMKEGG(self.org_code)])
            self._genematcher.set_targets(self.genes.keys())
        return self._genematcher
    genematcher = property(_get_genematcher, _set_genematcher)
    
    def get_genes(self):
        return self.genes
    
    @classmethod
    def organism_name_search(cls, name):
        genome = KEGGGenome()
        if name not in genome:
            ids = genome.search(name)
            if not ids:
                from . import obiTaxonomy
                ids = obiTaxonomy.search(name)
                ids = [id for id in ids if genome.search(id)]
            name = ids.pop(0) if ids else name
            
        try:
            return KEGGGenome()[name].entry_key()
        except KeyError:
            raise ValueError("Organims with name='%s' not found in KEGG." % name)
        
    @classmethod
    def organism_version(cls, name):
        name = cls.organism_name_search(name)
        orngServerFiles.localpath_download("KEGG", "kegg_genes_%s.tar.gz" % name)
        return orngServerFiles.info("KEGG", "kegg_genes_%s.tar.gz" % name)["datetime"]
              
class KEGGPathway(object):
    PNG_FILENAME = "%(path)s/pathway/organisms/%(org)s/%(org)s%(map_id)s.png"
    KGML_FILENAME = "%(path)s/xml/kgml/metabolic/organisms/%(org)s/%(org)s%(map_id)s.xml"
    VERSION = "v2.0"
    
    class entry(object):
        def __init__(self, dom_element):
            self.__dict__.update(dom_element.attributes.items())
            self.graphics = ()
            self.components = []
            self.graphics = dict(dom_element.getElementsByTagName("graphics")[0].attributes.items())
            self.components = [node.getAttribute("id") for node in dom_element.getElementsByTagName("component")]
    class reaction(object):
        def __init__(self, dom_element):
            self.__dict__.update(dom_element.attributes.items())
            self.substrates = [node.getAttribute("name") for node in dom_element.getElementsByTagName("substrate")]
            self.products = [node.getAttribute("name") for node in dom_element.getElementsByTagName("product")]
            
    class relation(object):
        def __init__(self, dom_element):
            self.__dict__.update(dom_element.attributes.items())
            self.subtypes = [node.attributes.items() for node in dom_element.getElementsByTagName("subtype")]
        
    def __init__(self, file):
        if not os.path.exists(file):
            path, org, map_id = self.split_pathway_id(file)
#            file = self.KGML_FILENAME % dict(path=DEFAULT_DATABASE_PATH, org=org, map_id=id)
            file = self.filename_kgml(org, map_id)
        if not os.path.exists(file) and os.path.exists(file.replace("metabolic", "non-metabolic")):
            file = file.replace("metabolic", "non-metabolic")
        self.filename = file
#        self.load(file)
        
    @persistent_cached_method
    def pathway_attributes(self):
        return dict(self.pathway_dom().attributes.items())
    
    @property
    def name(self):
        return self.pathway_attributes().get("name")
    
    @property
    def org(self):
        return self.pathway_attributes().get("org")
    
    @property
    def number(self):
        return self.pathway_attributes().get("number")
    
    @property    
    def title(self):
        return self.pathway_attributes().get("title")
    
    @property
    def image(self):
        return self.pathway_attributes().get("image")
    
    @property
    def link(self):
        return self.pathway_attributes().get("link")
    
    @cached_method
    @loads("KEGG", lambda self: "kegg_pathways_%s.tar.gz" % self.filename.split("/")[-1][:-9])
    def pathway_dom(self):
        return minidom.parse(self.filename).getElementsByTagName("pathway")[0]
    
    @cached_method
    def entrys(self):
        return [self.entry(e) for e in self.pathway_dom().getElementsByTagName("entry")]
    
    @cached_method
    def reactions(self):
        return [self.reaction(e) for e in self.pathway_dom().getElementsByTagName("reaction")]
    
    @cached_method
    def relations(self):
        return [self.relation(e) for e in self.pathway_dom().getElementsByTagName("relation")]
    
    @classmethod
    def split_pathway_id(cls, id):
        path, id = id.split(":") if ":" in id else ("path", id)
        org, id = id[:-5], id[-5:]
        return path, org, id  
        
    def __iter__(self):
        """ Iterate over all elements in the pathway
        """
        return iter(self.all_elements())
    
    def __contains__(self, element):
        """ Retrurn true if element in the pathway
        """
        return element in self.all_elements()
    
    def __getitem__(self, key):
        return
    
    @cached_method
    def all_elements(self):
        """ Return all elements
        """
        return reduce(list.__add__, [self.genes(), self.compounds(), self.enzmes(), self.reactions()], [])
    
    def _get_entrys_by_type(self, type):
        return reduce(set.union, [entry.name.split() for entry in self.entrys() if entry.type == type], set())
    
    @persistent_cached_method
    def genes(self):
        """ Return all genes on the pathway
        """
        return self._get_entrys_by_type("gene")
    
    @persistent_cached_method
    def compounds(self):
        """ Return all compounds on the pathway
        """
        return self._get_entrys_by_type("compound")
    
    @persistent_cached_method
    def enzymes(self):
        """ Return all enzymes on the pathway
        """
        return self._get_entrys_by_type("enzyme")
    
    @persistent_cached_method
    def orthologs(self):
        """ Return all orthologs on the pathway
        """
        return self._get_entrys_by_type("ortholog")
    
    @persistent_cached_method
    def maps(self):
        """ Return all linked maps on the pathway
        """
        return self._get_entrys_by_type("map")
    
    @persistent_cached_method
    def groups(self):
        """ Return all groups on the pathway
        """
        return self._get_entrys_by_type("ortholog")

    def get_image(self):
        """ Return an image of the pathway
        """
        return self.filename_png(self.org, self.number) % dict(path=DEFAULT_DATABASE_PATH)
    
    @persistent_cached_method
    def get_bounding_box_dict(self):
        return dict([(element.id, element.graphics) for element in self.entrys() if element.graphics])
    
    @persistent_cached_method
    def graphics(self, item):
        return [entry.graphics for entry in self.entrys() if item in entry.name and entry.graphics]
        
    @classmethod
    @partial(cached_method, cache_name="_cls_cached_method_cache_")
    @loads("KEGG", lambda cls, org: "kegg_pathways_%s.tar.gz" % org)
    def list(cls, org):
        file = cls.directory_png(org) + org + ".list"
        data = [line.split() for line in open(file, "rb").read().splitlines()]
        return reduce(lambda dict, line: dict[line[0]].update(line[1:]) or dict, data, defaultdict(set))
    
    @classmethod
    @defaultpath
    def filename_kgml(cls, org, map_id, path=DEFAULT_DATABASE_PATH):
        path = "%(path)s" if path is None else path
        if org in ["map", "ec"]:
            return "%(path)s/xml/kgml/metabolic/ec/ec%(map_id)s.xml" % dict(org=org, map_id=map_id, path=path)
        elif org == "ko":
            return "%(path)s/xml/kgml/metabolic/ko/ko%(map_id)s.xml" % dict(org=org, map_id=map_id, path=path)
        else:
            return "%(path)s/xml/kgml/metabolic/organisms/%(org)s/%(org)s%(map_id)s.xml" % dict(org=org, map_id=map_id, path=path)
        
    @classmethod
    @defaultpath
    def filename_png(cls, org, map_id, path=DEFAULT_DATABASE_PATH):
        path = "%(path)s" if path is None else path 
        if org in ["map", "ec", "ko", "rn"]:
            return "%(path)s/pathway/%(org)s/%(org)s%(map_id)s.png" % dict(org=org, map_id=map_id, path=path)
        else:
            return "%(path)s/pathway/organisms/%(org)s/%(org)s%(map_id)s.png" % dict(org=org, map_id=map_id, path=path)
        
    @classmethod
    @defaultpath
    def directory_png(cls, org, path=DEFAULT_DATABASE_PATH):
        return cls.filename_png(org, "", path=path).rsplit("/", 1)[0] + "/" 
    
    @classmethod
    @defaultpath
    def directory_kgml(cls, org, path=DEFAULT_DATABASE_PATH):
        return cls.filename_kgml(org, "", path=path).rsplit("/", 1)[0] + "/"
        
    @classmethod
    def pathways(cls, org):
        file = cls.directory_png(org) + org + ".list"
        pathways = [line.split()[0] for line in open(file, "rb").read().splitlines() if line.strip()]
        return sorted(set(pathways))
        
    @classmethod
    def download(cls, pathway, target=None):
        """ Download the pathway (xml and png files) to target directory
        """
        org = pathway[:3]
        import urllib2
        xml_data = urllib2.urlopen("ftp://ftp.genome.jp/pub/kegg/xml/kgml/metabolic/organisms/%s/%s.xml" % (org, pathway)).read()
        open(os.path.join(target, pathway + ".xml"), "wb").write(xml_data)
        png_data = urllib2.urlopen("ftp://ftp.genome.jp/pub/kegg/pathway/organisms/%s/%s.png" % (org, pathway)).read()
        open(os.path.join(target, pathway + ".png"), "wb").write(png_data)
        
    @classmethod
    @downloader
    def download_pathways(cls, org):
        png_path = cls.directory_png(org, path=None).replace("%(path)s/", "")
        xml_path = cls.directory_kgml(org, path=None).replace("%(path)s/", "")
        data = urllib2.urlopen("ftp://ftp.genome.jp/pub/kegg/" + png_path + org + ".list").read()
        pathways = sorted(set([line.split()[0][5:] for line in data.splitlines() if line]))
        from .obiData import FtpDownloader
        ftp = FtpDownloader("ftp.genome.jp", _join(), "pub/kegg/", numOfThreads=15)
        ftp.massRetrieve([png_path + pathway + ".png" for pathway in pathways])
            
        if org != "map" :
            pathways = ftp.listdir(xml_path)
            pathways = [p for p in pathways if p.endswith(".xml")]
            ftp.massRetrieve([xml_path + pathway for pathway in pathways])
            if org not in ["ec", "rn"]: # Add non-metabolic pathways to the download list
                xml_path = xml_path.replace("metabolic", "non-metabolic")
                pathways = ftp.listdir(xml_path)
                pathways = [p for p in pathways if p.endswith(".xml")] 
                ftp.massRetrieve([xml_path + pathway for pathway in pathways])
        ftp.massRetrieve([png_path + org + ".list"])
        return []
    
persistent_cached_class(KEGGPathway)
        
def organism_name_search(name):
    return KEGGOrganism.organism_name_search(name)

def pathways(org):
    return KEGGPathway.list(org)

def organisms():
    return KEGGOrganism.organisms()

def to_taxid(name):
    genome = KEGGGenome()
    names = genome.search(name)
    if genome[names[0]].taxonomy:
        return re.findall(r"TAX:\s*(\d+)", genome[names[0]].taxonomy)[0]
    else:
        return None #Should this raise an error?

def from_taxid(taxid):
    return KEGGGenome().search(taxid)[0]
    
def test():
    p = KEGGPathway("sce00010.xml")
    print p.genes
    print p.reactions
    print p.compounds
    print p.image
    g = KEGGGenome()
    org = KEGGOrganism("Homo sapiens")
    print list(org.genes)[:10]
    org.gene_aliases
    print org.pathways(with_ids=org.genes.keys()[:5])
    print org.enzymes()
    print org.enriched_pathways(org.genes.keys()[:10])
    print org.genematcher

if __name__=="__main__":
    test()
#    org1 = KEGGOrganism("ddi")
#    org2 = KEGGOrganism("ddi")
#    org2.api = KEGGInterface()
#    tests = [("get_genes", ()),
#             ("get_genes_by_enzyme", ("ec:1.1.1.1",)),
#             ("get_genes_by_pathway", ("path:ddi00010",)),
#             ("get_pathways_by_genes", (["ddi:DDB_0191256"],)),
#             ("get_pathways_by_enzymes", (["ec:1.1.1.1"],)),
#             ("get_pathways_by_compounds", (["cpd:C00001"],)),
#             ("get_linked_pathways", ("path:ddi00010",)),
#             ("list_pathways", ()),
#             ("get_compounds_by_enzyme", ("ec:1.1.1.1",)),
#             ("get_compounds_by_pathway", ("path:ddi00010",)),
#             ("get_enzymes_by_compound", ("cpd:C00001",)),
#             ("get_enzymes_by_pathway", ("path:ddi00010",)),
#             ("get_enzymes_by_gene", ("ddi:DDB_0191256",))]
#    for name, args in tests:
#        s1 = set(getattr(org1, name)(*args))
#        s2 = set(getattr(org2, name)(*args))
#        if s1 and s2:
#            print name
#            print s1-s2
#            print s2-s1
#        else:
#            print name
#            print "both empty"



"""\
obiOntology
-----------

Python module for manipulating OBO Ontology files (http://www.obofoundry.org/)

You can construct an ontology from scratch with custom terms ::

    >>> term = OBOObject("Term", id="foo:bar", name="Foo bar")
    >>> print term
    [Term]
    id: foo:bar
    name: Foo bar
    
    >>> ontology = OBOOntology()
    >>> ontology.add_object(term)
    >>> ontology.add_header_tag("created-by", "ales") # add a header tag
    >>> from StringIO import StringIO
    >>> buffer = StringIO()
    >>> ontology.dump(buffer) # Save the ontology to a file like object
    >>> print buffer.getvalue() # Print the contents of the buffer
    created-by: ales
    <BLANKLINE>
    [Term]
    id: foo:bar
    name: Foo bar
    <BLANKLINE>
    
To load an ontology from a file pass the file or filename to the OBOOntology
constructor or call its load method ::

    >>> buffer.seek(0) # rewind
    >>> ontology = OBOOntology(buffer)
    >>> # Or
    >>> buffer.seek(0) # rewind
    >>> ontology = OBOOntology()
    >>> ontology.load(buffer)
    
See OBOOntology docstring for more details.

"""

from itertools import chain
from collections import defaultdict
from StringIO import StringIO
import itertools
import re

BUILTIN_OBO_OBJECTS = [
"""[Typedef]
id: is_a
name: is_a
range: OBO:TERM_OR_TYPE
domain: OBO:TERM_OR_TYPE
definition: The basic subclassing relationship [OBO:defs]"""
,
"""[Typedef]
id: disjoint_from
name: disjoint_from
range: OBO:TERM
domain: OBO:TERM
definition: Indicates that two classes are disjoint [OBO:defs]"""
,
"""[Typedef]
id: instance_of
name: instance_of
range: OBO:TERM
domain: OBO:INSTANCE
definition: Indicates the type of an instance [OBO:defs]"""
,
"""[Typedef]
id: inverse_of
name: inverse_of
range: OBO:TYPE
domain: OBO:TYPE
definition: Indicates that one relationship type is the inverse of another [OBO:defs]"""
,
"""[Typedef]
id: union_of
name: union_of
range: OBO:TERM
domain: OBO:TERM
definition: Indicates that a term is the union of several others [OBO:defs]"""
,
"""[Typedef]
id: intersection_of
name: intersection_of
range: OBO:TERM
domain: OBO:TERM
definition: Indicates that a term is the intersection of several others [OBO:defs]"""
]
    
def _split_and_strip(string, sep):
    head, tail = string.split(sep, 1)
    return head.rstrip(" "), tail.lstrip(" ")


class OBOObject(object):
    """ Represents a generic OBO object (e.g. Term, Typedef, Instance, ...)
    Example::
        >>> term = OBOObject(stanza_type="Term", id="FOO:001", name="bar")
    """
    def __init__(self, stanza_type="Term", **kwargs):
        """ Init from keyword arguments.
        Example::
            >>> term = OBOObject(stanza_type="Term", id="FOO:001", name="bar", def_="Example definition { modifier=frob } ! Comment")
            >>> term = OBOObject(stanza_type="Term", id="FOO:001", name="bar", def_=("Example definition", [("modifier", "frob")], "Comment"))
            >>> term = OBOObject(stanza_type="Term", id="FOO:001", name="bar", def_=("Example definition", [("modifier", "frob")])) # without the comment
            >>> term = OBOObject(stanza_type="Term", id="FOO:001", name="bar", def_=("Example definition",)) # without the modifiers and comment
        """
        self.stanza_type = stanza_type
        
        self.modifiers = []
        self.comments = []
        self.tag_values = []
        self.values = {}
        
        sorted_tags = sorted(kwargs.iteritems(), key=lambda key_val: chr(1) if key_val[0] == "id" else key_val[0])
        for tag, value in sorted_tags:
            if isinstance(value, basestring):
                tag, value, modifiers, comment = self.parse_tag_value(self.name_demangle(tag), value)
            elif isinstance(value, tuple):
                tag, value, modifiers, comment = ((self.name_demangle(tag),) + value + (None, None))[:4]
            self.add_tag(tag, value, modifiers, comment)
        
        self.related = set()
#        self.related_to = set()
            
    @property
    def is_annonymous(self):
        value = self.get_value("is_annonymous")
        return bool(value)
    
    def name_mangle(self, tag):
        """ Mangle tag name if it conflicts with python keyword
        Example::
            >>> term.name_mangle("def"), term.name_mangle("class")
            ('def_', 'class_')
        """
        if tag in ["def", "class", "in", "not"]:
            return tag + "_"
        else:
            return tag
        
    def name_demangle(self, tag):
        """ Reverse of name_mangle
        """
        if tag in ["def_", "class_", "in_", "not_"]:
            return tag[:-1]
        else:
            return tag
        
    def add_tag(self, tag, value, modifiers=None, comment=None):
        """ Add `tag`, `value` pair to the object with optional modifiers and
        comment.
        Example::
            >>> term = OBOObject("Term")
            >>> term.add_tag("id", "FOO:002", comment="This is an id")
            >>> print term
            [Term]
            id: FOO:002 ! This is an id
             
        """
        tag = intern(tag) # a small speed and memory benefit
        self.tag_values.append((tag, value))
        self.modifiers.append(modifiers)
        self.comments.append(comment)
        self.values.setdefault(tag, []).append(value)
        
        #  TODO: fix multiple tags grouping
        if hasattr(self, tag):
            if isinstance(getattr(self, tag), list):
                getattr(self, tag).append(value)
            else:
                setattr(self, tag, [getattr(self, tag)] + [value])
        else:
            setattr(self, self.name_mangle(tag), value)
            
    def update(self, other):
        """ Update the term with tag value pairs from `other` 
        (a OBOObject instance). The tag value pairs are appended
        to the end except for the `id` tag.
        """ 
        for (tag, value), modifiers, comment in zip(other.tag_values, other.modifiers, other.comments):
            if tag != "id":
                self.add_tag(tag, value, modifiers, comment)
        
    def get_value(self, tag, group=True):
        if group:
            pairs = [pair for pair in self.tag_values if pair[0] == tag]
            return pairs
        else:
            tag = self.name_mangle(tag)
            if tag in self.__dict__:
                return self.__dict__[tag]
            else:
                raise ValueError("No value for tag: %s" % tag)
        
    def tag_count(self):
        """ Retrun the number of tags in this object
        """
        return len(self.tag_values)
    
    def tags(self):
        """ Retrun an iterator over the (tag, value) pairs.
        """
        for i in range(self.tag_count()):
            yield self.tag_values[i] + (self.modifiers[i], self.comments[i])
        
    def format_single_tag(self, index):
        """Return a formated string representing index-th tag pair value
        Example::
            >>> term = OBOObject("Term", id="FOO:001", name="bar", def_="Example definition {modifier=frob} ! Comment")
            >>> term.format_single_tag(0)
            'id: FOO:001'
            >>> term.format_single_tag(1)
            'def: Example definition { modifier=frob } ! Comment'
        """
        tag, value = self.tag_values[index]
        modifiers = self.modifiers[index]
        comment = self.comments[index]
        res = ["%s: %s" % (tag, value)]
        if modifiers:
            res.append("{ %s }" % modifiers)
        if comment:
            res.append("! " + comment)
        return " ".join(res)
    
    def format_stanza(self):
        """ Return a string stanza representation of this object 
        """
        stanza = ["[%s]" % self.stanza_type]
        for i in range(self.tag_count()):
            stanza.append(self.format_single_tag(i))
        return "\n".join(stanza)
            
    @classmethod     
    def parse_stanza(cls, stanza):
        r""" Parse and return an OBOObject instance from a single stanza.
        Example::
            >>> term = OBOObject.parse_stanza("[Term]\nid: FOO:001\nname:bar")
            >>> print term.id, term.name
            FOO:001 bar
            
        """
        lines = stanza.splitlines()
        stanza_type = lines[0].strip("[]")
        tag_values = []
        for line in lines[1:]:
            if ":" in line:
                tag_values.append(cls.parse_tag_value(line))
        
        obo = OBOObject(stanza_type)
        for i, (tag, value, modifiers, comment) in enumerate(tag_values):
#            print tag, value, modifiers, comment
            obo.add_tag(tag, value, modifiers, comment)
        return obo
    
        
    @classmethod
    def parse_tag_value_1(cls, tag_value_pair, *args):
        """ Parse and return a four-tuple containing a tag, value, a list of modifier pairs, comment.
        If no modifiers or comments are present the corresponding entries will be None.
        
        Example::
            >>> OBOObject.parse_tag_value("foo: bar {modifier=frob} ! Comment")
            ('foo', 'bar', 'modifier=frob', 'Comment')
            >>> OBOObject.parse_tag_value("foo: bar")
            ('foo', 'bar', None, None)
            >>> #  Can also pass tag, value pair already split   
            >>> OBOObject.parse_tag_value("foo", "bar {modifier=frob} ! Comment")
            ('foo', 'bar', 'modifier=frob', 'Comment')
        """
        if args and ":" not in tag_value_pair:
            tag, rest = tag_value_pair, args[0]
        else:
            tag, rest = _split_and_strip(tag_value_pair, ":")
        value, modifiers, comment = None, None, None
        
        if "{" in rest:
            value, rest = _split_and_strip(rest, "{",)
            modifiers, rest = _split_and_strip(rest, "}")
        if "!" in rest:
            if value is None:
                value, comment = _split_and_strip(rest, "!")
            else:
                _, comment = _split_and_strip(rest, "!")
        if value is None:
            value = rest
            
        if modifiers is not None:
            modifiers = modifiers #TODO: split modifiers in a list
            
        return tag, value, modifiers, comment
    
    _RE_TAG_VALUE = re.compile(r"^(?P<tag>.+?[^\\])\s*:\s*(?P<value>.+?)\s*(?P<modifiers>[^\\]{.+?[^\\]})?\s*(?P<comment>[^\\]!.*)?$")
    _RE_VALUE = re.compile(r"^\s*(?P<value>.+?)\s*(?P<modifiers>[^\\]{.+?[^\\]})?\s*(?P<comment>[^\\]!.*)?$")
    
    @classmethod
    def parse_tag_value(cls, tag_value_pair, arg=None):
        """ Parse and return a four-tuple containing a tag, value, a list of modifier pairs, comment.
        If no modifiers or comments are present the corresponding entries will be None.
        
        Example::
            >>> OBOObject.parse_tag_value("foo: bar {modifier=frob} ! Comment")
            ('foo', 'bar', 'modifier=frob', 'Comment')
            >>> OBOObject.parse_tag_value("foo: bar")
            ('foo', 'bar', None, None)
            >>> #  Can also pass tag, value pair already split   
            >>> OBOObject.parse_tag_value("foo", "bar {modifier=frob} ! Comment")
            ('foo', 'bar', 'modifier=frob', 'Comment')
            
        .. warning: This function assumes comment an modifiers are prefixed
            with a whitespace i.e. 'tag: bla! comment' will be parsed incorrectly!
        """
        if arg is not None: # tag_value_pair is actually a tag only
            tag = tag_value_pair
            value, modifiers, comment =  cls._RE_VALUE.findall(arg)[0]
        else:
            tag, value, modifiers, comment = cls._RE_TAG_VALUE.findall(tag_value_pair)[0]
        none_if_empyt = lambda val: None if not val.strip() else val.strip()
        modifiers = modifiers.strip(" {}")
        comment = comment.lstrip(" !")
        return (none_if_empyt(tag), none_if_empyt(value),
                none_if_empyt(modifiers), none_if_empyt(comment))
         
    def related_objects(self):
        """ Return a list of tuple pairs where the first element is relationship (typedef id)
        is and the second object id whom the relationship applies to.
        """
        result = [(type_id, id) for type_id in ["is_a"] for id in self.values.get(type_id, [])] ##TODO add other defined Typedef ids
        result = result + [tuple(r.split(None, 1)) for r in self.values.get("relationship", [])]
        return result

    def __repr__(self):
        """ Return a string representation of the object in OBO format
        """
        return self.format_stanza()

    def __iter__(self):
        """ Iterates over sub terms
        """
        for type_id, id in self.related_objects():
            yield (type_id, id)
        
        
class Term(OBOObject):
    def __init__(self, *args, **kwargs):
        OBOObject.__init__(self, "Term", *args, **kwargs)

class Typedef(OBOObject):
    def __init__(self, *args, **kwargs):
        OBOObject.__init__(self, "Typedef", *args, **kwargs)

class Instance(OBOObject):
    def __init__(self, *args, **kwargs):
        OBOObject.__init__(self, "Instance", *args, **kwargs)


import re

class OBOParser(object):
    r""" A simple parser for .obo files (inspired by xml.dom.pulldom)
    
    Example::
        >>> from StringIO import StringIO
        >>> file = StringIO("header_tag: header_value\n[Term]\nid: FOO { modifier=bar } ! comment\n\n") 
        >>> parser = OBOParser(file)
        >>> for event, value in parser:
        ...     print event, value
        ...     
        HEADER_TAG ['header_tag', 'header_value']
        START_STANZA Term
        TAG_VALUE ('id', 'FOO', 'modifier=bar', 'comment')
        CLOSE_STANZA None
                
    """
    def __init__(self, file):
        self.file = file
        
    def parse(self, progress_callback=None):
        """ Parse the file and yield parse events.
        
        .. TODO: List events and values
        """
        data = self.file.read()
        header = data[: data.index("\n[")]
        body = data[data.index("\n[") + 1:]
        for line in header.splitlines():
            if line.strip():
                yield "HEADER_TAG", line.split(": ", 1)
                
        current = None
        #  For speed make these functions local
        startswith = str.startswith
        endswith = str.endswith
        parse_tag_value = OBOObject.parse_tag_value
        
        for line in body.splitlines():
#            line = line.strip()
            if startswith(line, "[") and endswith(line, "]"):
                yield "START_STANZA", line.strip("[]")
                current = line
            elif startswith(line, "!"):
                yield "COMMENT", line[1:]
            elif line:
                yield "TAG_VALUE", parse_tag_value(line)
            else: #  empty line is the end of a term
                yield "CLOSE_STANZA", None
                current = None
        if current is not None:
            yield "CLOSE_STANZA", None
    
    def __iter__(self):
        """ Iterate over parse events (same as parse())
        """
        return self.parse() 
        
        
class OBOOntology(object):
    """ The main ontology object.
    """
    
    BUILTINS = BUILTIN_OBO_OBJECTS
    def __init__(self, file=None):
        """ Init an ontology instance from a file like object (.obo format)
        
        """
        self.objects = []
        self.header_tags = []
        self.id2term = {}
        self.alt2id = {}
        self._resolved_imports = []
        self._invalid_cache_flag = False
        self._related_to = {}
        
        # First load the built in OBO objects
        builtins = StringIO("\n" + "\n\n".join(self.BUILTINS) + "\n")
        self.load(builtins)
        if file:
            self.load(file)
        
    def add_object(self, object):
        """ Add OBOObject instance to this ontology.
        """
        if object.id in self.id2term:
            raise ValueError("OBOObject with id: %s already in the ontology" % object.id)
        self.objects.append(object)
        self.id2term[object.id] = object
        self._invalid_cache_flag = True
        
    def add_header_tag(self, tag, value):
        """ Add header tag, value pair to this ontology
        """
        self.header_tags.append((tag, value))
    
    def load(self, file, progress_callback=None):
        """ Load terms from a file.
        """
        if isinstance(file, basestring):
            file = open(file, "rb")
        parser = OBOParser(file)
        current = None
        for event, value in parser.parse(progress_callback=progress_callback):
            if event == "TAG_VALUE":
                current.add_tag(*value)
            elif event == "START_STANZA":
                current = OBOObject(value)
            elif event == "CLOSE_STANZA":
                self.add_object(current)
                current = None
            elif event == "HEADER_TAG":
                self.add_header_tag(*value)
            elif event != "COMMENT":
                raise Exception("Parse Error! Unknown parse event {0}".format(event))
            
        imports = [value for tag, value, in self.header_tags if tag == "import"]
        
        while imports:
            url = imports.pop(0)
            if uri not in self._resolved_imports:
                imported = self.parse_file(open(url, "rb"))
                ontology.update(imported)
                self._resolved_imports.append(uri)
                
    def dump(self, file):
        """ Dump the contents of the ontology to a .obo `file`.
        """
        if isinstance(file, basestring):
            file = open(file, "wb")
            
        for key, value in self.header_tags:
            file.write(key + ": " + value + "\n")
            
        # Skip the builtins
        for object in self.objects[len(self.BUILTINS):]:
            file.write("\n")
            file.write(object.format_stanza())
            file.write("\n")
    
    def update(self, other):
        """ Update this ontology with the terms from another. 
        """
        for term in other:
            if term.id in self:
                if not term.is_annonymous:
                    self.term(term.id).update(term)
                else: #  Do nothing
                    pass 
            else:
                self.add_object(term)
        self._invalid_cache_flag = True
        
    def _cache_validate(self, force=False):
        """ Update the relations cache if `self._invalid_cache` flag is set. 
        """
        if self._invalid_cache_flag or force:
            self._cache_relations()
            
    def _cache_relations(self):
        """ Collect all relations from parent to a child and store it in
        `self._related_to` member.
        
        """
        related_to = defaultdict(list)
        for obj in self.objects:
            for rel_type, id in self.related_terms(obj):
                term = self.term(id)
                related_to[term].append((rel_type, obj))
                
        self._related_to = related_to
        self._invalid_cache_flag = False
        
    def term(self, id):
        """ Return the OBOObject associated with this id.
        """
        if isinstance(id, basestring):
            if id in self.id2term:
                return self.id2term[id]
            elif id in self.alt2id:
                return self.id2term[self.alt2id[id]]
            else:
                raise ValueError("Unknown term id: %r" % id)
        elif isinstance(id, OBOObject):
            return id
        
    def terms(self):
        """ Return all `Term` instances in the ontology.
        """
        return [obj for obj in self.objects if obj.stanza_type == "Term"]
    
    def term_by_name(self, name):
        """ Return the term with name ``name``. 
        """
        terms = [t for t in self.terms() if t.name == name]
        if len(terms) != 1:
            raise ValueError("Unknown term name: %r" % name)
        return terms[0]
    
    def typedefs(self):
        """ Return all `Typedef` instances in the ontology.
        """
        return [obj for obj in self.objects if obj.stanza_type == "Typedef"]
    
    def instances(self):
        """ Return all `Instance` instances in the ontology.
        """
        return [obj for obj in self.objects if obj.stanza_type == "Instance"]
        
    def related_terms(self, term):
        """ Return a list of (rel_type, term_id) tuples where rel_type is
        relationship type (e.g. 'is_a', 'has_part', ...) and term_id is the
        id of the term in the relationship.
        
        """
        term = self.term(term) if not isinstance(term, OBOObject) else term
        related = [(tag, value) for tag in ["is_a"] for value in term.values.get(tag, [])] #TODO: add other typedef ids
        relationships = term.values.get("relationship", [])
        for rel in relationships:
            related.append(tuple(rel.split(None, 1)))
        return related
        
    def edge_types(self):
        """ Return a list of all edge types in the ontology
        """
        return [obj.id for obj in self.objects if obj.stanza_type == "Typedef"]
    
    def parent_edges(self, term):
        """ Return a list of (rel_type, parent_term) tuples 
        """
        term = self.term(term)
        parents = []
        for rel_type, parent in self.related_terms(term):
            parents.append((rel_type, self.term(parent)))
        return parents
        
    def child_edges(self, term):
        """ Return a list of (rel_type, source_term) tuples
        """
        self._cache_validate()
        term = self.term(term)
        return self._related_to.get(term, [])
        
        
    def super_terms(self, term):
        """ Return a set of all super terms of `term` up to the most general one.
        """
        terms = self.parent_terms(term)
        visited = set()
        queue = set(terms)
        while queue:
            term = queue.pop()
            visited.add(term)
            queue.update(self.parent_terms(term) - visited)
        return visited
    
    def sub_terms(self, term):
        """ Return a set of all sub terms for `term`.
        """
        terms = self.child_terms(term)
        visited = set()
        queue = set(terms)
        while queue:
            term = queue.pop()
            visited.add(term)
            queue.update(self.child_terms(term) - visited)
        return visited
    
    def child_terms(self, term):
        """ Return a set of all child terms for this `term`.
        """
        self._cache_validate()
        term = self.term(term)
        children = []
        for rel_type, term in self.child_edges(term):
            children.append(term)
        return set(children)
        
    def parent_terms(self, term):
        """ Return a set of all parent terms for this `term`
        """
        term = self.term(term)
        parents = []
        for rel_type, id in self.parent_edges(term): #term.related_objects():
            parents.append(self.term(id))
        return set(parents)
    
    def relations(self):
        """ Return a list of all relations in the ontology.
        """
        relations = []
        for obj in self.objects:
            for type_id, id in  obj.related:
                target_term = self.term(id)
            relations.append((obj, type_id, target_term))
        return relations
    
    def __len__(self):
        return len(self.objects)
    
    def __iter__(self):
        return iter(self.objects)
    
    def __contains__(self, obj):
        if isinstance(obj, basestring):
            return obj in self.id2term
        else:
            return obj in self.objects
    
    def __getitem__(self, key):
        return self.id2term[key]
    
    def has_key(self, key):
        return self.id2tem.has_key(key)
    
    def traverse_bf(self, term):
        """ BF traverse of the ontology down from term.
        """
        queue = list(self.child_terms(term))
        while queue:
            term = queue.pop(0)
            queue.extend(self.child_terms(term))
            yield term
                    
    def traverse_df(self, term, depth=1e30):
        """ DF traverse of the ontology down from term.
        """
        if depth >= 1:
            for child in self.child_terms(term):
                yield child
                for t in self.traverse_df(child, depth-1):
                    yield t
        
    
    def to_network(self, terms=None):
        """ Return an Orange.network.Network instance constructed from
        this ontology.
        
        """
        edge_types = self.edge_types()
        terms = self.terms()
        from Orange.orng import orngNetwork
        import orange
        
        network = orngNetwork.Network(len(terms), True, len(edge_types))
        network.objects = dict([(term.id, i) for i, term in enumerate(terms)])
        
        edges = defaultdict(set)
        for term in self.terms():
            related = self.related_terms(term)
            for relType, relTerm in related:
                edges[(term.id, relTerm)].add(relType)
                
        edgeitems = edges.items()
        for (src, dst), eTypes in edgeitems:
            network[src, dst] = [1 if e in eTypes else 0 for e in edge_types]
            
        domain = orange.Domain([orange.StringVariable("id"),
                                orange.StringVariable("name"),
                                orange.StringVariable("def"),
                                ], False)
        
        items = orange.ExampleTable(domain)
        for term in terms:
            ex = orange.Example(domain, [term.id, term.name, term.values.get("def", [""])[0]])
            items.append(ex)
        
        relationships = set([", ".join(sorted(eTypes)) for (_, _), eTypes in edgeitems])
        domain = orange.Domain([orange.FloatVariable("u"),
                                orange.FloatVariable("v"),
                                orange.EnumVariable("relationship", values=list(edge_types))
                                ], False)
        
        id2index = dict([(term.id, i + 1) for i, term in enumerate(terms)])
        links = orange.ExampleTable(domain)
        for (src, dst), eTypes in edgeitems:
            ex = orange.Example(domain, [id2index[src], id2index[dst], eTypes.pop()])
            links.append(ex)
            
        network.items = items
        network.links = links
        network.optimization = None
        return network
    
    def to_networkx(self, terms=None):
        """ Return a NetworkX graph of this ontology
        """
        import networkx
        graph = networkx.Graph()
        
        edge_types = self.edge_types()
        
        edge_colors = {"is_a": "red"}
        
        if terms is None:
            terms = self.terms()
        else:
            terms = [self.term(term) for term in terms]
            super_terms = [self.super_terms(term) for term in terms]
            terms = reduce(set.union, super_terms, set(terms))
            
        for term in terms:
            graph.add_node(term.id, name=term.name)
            
        for term in terms:
            for rel_type, rel_term in self.related_terms(term):
                rel_term = self.term(rel_term)
                if rel_term in terms:
                    graph.add_edge(term.id, rel_term.id, label=rel_type, color=edge_colors.get(rel_type, "blue"))
                    
        return graph
    
    def to_graphviz(self, terms=None):
        """ Return an pygraphviz.AGraph representation of the ontology in.
        If `terms` is not `None` it must be a list of terms in the ontology.
        The graph will in this case contain only the super graph of those
        terms.  
        
        """
        import pygraphviz as pgv
        graph = pgv.AGraph(directed=True, name="ontology")
        
        edge_types = self.edge_types()
        
        edge_colors = {"is_a": "red"}
        
        if terms is None:
            terms = self.terms()
        else:
            terms = [self.term(term) for term in terms]
            super_terms = [self.super_terms(term) for term in terms]
            terms = reduce(set.union, super_terms, set(terms))
            
        for term in terms:
            graph.add_node(term.id, name=term.name)
            
        for term in terms:
            for rel_type, rel_term in self.related_terms(term):
                rel_term = self.term(rel_term)
                if rel_term in terms:
                    graph.add_edge(term.id, rel_term.id, label=rel_type, color=edge_colors.get(rel_type, "blue"))
                    
        return graph
    
    
def load(file):
    """ Load an ontology from a .obo file
    """
    return OBOOntology(file)
    
    
def foundry_ontologies():
    """ List ontologies available from the OBOFoundry website
    (`http://www.obofoundry.org/`_) 
    Example::
        >>> foundry_ontologies()
        [('Biological process', 'http://obo.cvs.sourceforge.net/*checkout*/obo/obo/ontology/genomic-proteomic/gene_ontology_edit.obo'), ...
    
    """
    import urllib2, re
    stream = urllib2.urlopen("http://www.obofoundry.org/")
    text = stream.read()
    pattern = r'<td class=".+?">\s*<a href=".+?">(.+?)</a>\s*</td>\s*<td class=".+?">.*?</td>\s*<td class=".+?">.*?</td>\s*?<td class=".+?">\s*<a href="(.+?obo)">.+?</a>'
    return re.findall(pattern, text)
    
    
if __name__ == "__main__":
    import doctest
    stanza = '''[Term]
id: FOO:001
name: bar
'''
    from StringIO import StringIO
    seinfeld = StringIO("""
[Typedef]
id: parent

[Typedef]
id: child
inverse_of: parent ! not actually used yet

[Term]
id: 001
name: George

[Term]
id: 002
name: Estelle
relationship: parent 001 ! George

[Term]
id: 003
name: Frank
relationship: parent 001 ! George

""") # TODO: fill the ontology with all characters
    term = OBOObject.parse_stanza(stanza)
    
    seinfeld = OBOOntology(seinfeld)
    print seinfeld.child_edges("001")
    
    doctest.testmod(extraglobs={"stanza": stanza, "term": term}, optionflags=doctest.ELLIPSIS)
        
